; editor tab setting: 8
;
;                           D                88888         //
;                           D               8     8       //
;                           D               8     8      //    EEEEE
;           PPPPPP     DDDDDD    PPPPPP      88888      //    E     E
;           P     P   D     D    P     P    8     8    //     EEEEEE
;           P     P   D     D    P     P    8     8   //      E
;           PPPPPP     DDDDDD    PPPPPP      88888   //        EEEEE
;           P                    P
;           P                    P
;           P                    P
;
;
;                              MEMORY ADDRESS
;-------------------------------------------------------------          -------
;|    EMA    |           |           |           |           |          | RUN |
;-------------------------------------------------------------          -------
;  O   O   O | O   O   O | O   O   O | O   O   O | O   O   O |             O
;            |           |           |           |           |
;            | O   O   O | O   O   O | O   O   O | O   O   O |
;            |-----------|-----------|-----------|-----------|
;            |   |   |   |   |   |   |   |   |   |   |   |   |STATE -----
;            |-----------|-----------|-----------|-----------|           \
;            |   |   |   |   |   |   |   |   |   |   |   |   |STATUS ---  \
;            |-----------|-----------|-----------|-----------|          \ _
;            |   |   |   |   |   |   |   |   |   |   |   |   | AC  ----  /  \
;            |   |   |   |   |   |   |   |   |   |   |   |   | MD  ---  |o   |
;            |   |   |   |   |   |   |   |   |   |   |   |   | MQ  ----  \_ /
;            |___|___|___|___|___|___|___|___|___|___|___|___| BUS ----/  
;
;     SW       0   1   2   3   4   5   6   7   8   9  10  11
;    |--|    |---+---+---+---+---+---+---+---+---+---+---+---|   |--+--|  |--+- - - -
;    |==|    |===|===|===|===|===|===|===|===|===|===|===|===|   |==|==|  |==|= : : :
;    |--|    |---+---+---+---+---+---+---+---+---+---+---+---|   |--+--+  |--+- - - -
;
;
;  SWITCHES and KEYS 
; -------------------
;  [s] SW		- when this switch is up, the line on the OMNIBUS called SW
;			  is high; when the switch is down, the line is low. This
;			  switch is used by special peripheral controls, for example
;			  the Bootstrap Loader.
;  [s] SWITCH REGISTER	- these 12 switches provide a means of communication between
;	(12 switches)	  operator and machine. They allow a 12-bit word to be input.
;			  When the switch is up is designates a binary 1, switch down
;			  is a binary 0. These switches are used during manual functs
;			  or under program control.
;
;
;            ADDR EXTD        START   EXAM HALT SING      DEP
;            LOAD ADDR     CLEAR CONT           STEP
;                 LOAD
;           |----+----+    +----+----+----+----+----+    +----+
;           |====|====|    |====|====|====|====|====|    |====|
;           |----|----+    +----+----+----+----+----+    +----+
;
;
;  [k] ADDR LOAD	- this key loads the contents of the Switch Register into the
;			  PC and forces Fetch to be set (no Major States while the
;			  Load Address Key is depressed).
;  [k] EXTD ADDR LOAD	- this key loads the contents of SR6-11 into the Data Field and
;			  Instruction Field registers of the Memory Extension Control.
;			  SR9-11 goes to Data Field 0-1-2. SR6-8 goes to Instruction
;			  Field 0-1-2.
;  [k] CLEAR		- this key issues an Initialize Pulse, clearing the AC, LINK,
;			  Interrupt system, and I/O Flags.
;  [k] CONT		- this key resumes the computer program by issueing a Memory
;			  Start and setting the Run flip-flop. The word stored at the
;			  address currently held by the PC is taken as the first
;			  instruction.
;  [k] EXAM		- puts the contents of the core memory at the address specified
;			  by the contents of the PC into the MB and PC. Then the
;			  contents of the PC is incremented by one to allow examination
;			  of the contents of sequential core memory addresses by
;			  repeating the operation of the EXAMine key.
;  [s] HALT		- this switch clears the Run flip-flop and causes the machine
;			  to stop at TS1 of the next Fetch cycle. This switch is also
;			  used for single instruction stepping. HALT is active when the
;			  switch is in the "down" position.
;  [s] SING STEP	- this switch clears the Run flip-flop and causes the machine
;			  to stop at TS1 of the next Fetch cycle. Thereafter, repeated
;			  depressing of the CONTinue keys steps the program one cycle
;			  at a time, so that the contents of registers can be observed
;			  in each state. SING STEP is active when the switch is in the
;			  "down" position.
;  [k] DEP		- loads the contents of the SR into the MB and core memory at
;			  the address given by the current contents of the PC. Then the
;			  contents of the PC is incremented by one. This allows storing
;			  of information in sequential memory addresses by repeated
;			  operation of the DEPosit key.
;
; There are two modes "single step", and "single instruction".  Each is set by pushing
; down the appropriate switch. IIRC, single instruction mode stops before(?) the fetch
; cycle of each instruction. (That would be after the execute cycle of the previous
; instruction.)
; Singe *step* mode stops after the fetch, after the defer, and after the execute.
; Not sure how it interacts with data break and other special cycles.
;
;
;  LAMPS
; -------
;  Memory Address  indicates the contents of the memory address that will be accessed
;		   next.
;  EMA		   indicates which Extended Memory Field is being accessed
;		   IF0-2 (in STATUS) are the three EMA bits for a fetch or a direct
;		   memory reference. DF0-2 are the three EMA bits for an indirect
;		   memory reference. The three EMA bits control the actual memory
;		   reference being performed at any given instant.  "F" and "D" cycles
;		   always IF, "E" cycle memory accesses depend on whether the previous
;		   cycle was a "D" or not.
;  RUN		   when lit means the machine's timing is enabled and capable of
;		   executing instructions.
;
;   _________________________________________________________________
;   |  F    D    E   IR0  IR1  IR2  MD   DATA  SW  PAUSE  BRK   BRK  |
;   |                               DIR  CONT             PROG       | STATE
;   +-==============================================================-+
;   | LNK  GT   INT   NO  ION   UM  IF0  IF1   IF2  DF0   DF1   DF2  |
;   |           BUS  INT                                             | STATUS
;   +-==============================================================-+
;   |  0    1    2    3    4    5    6    7    8    9    10     11   | AC,MD,MQ,BUS
;   ------------------------------------------------------------------
;
;   note: only "PAUSE", "BRK PROG", "BRK", and "GT" are not implemented at this point.
;
;
;  INDICATOR SELECTOR SWITCH (ISS)
; ====================================================================================
;
;  STATE (STE)	- indicates an individual function for each bit:
;
;   0 - Fetch	  is ON for a FETCH cycle, reading an instruction from memory
;   1 - Defer	  is ON for a DEFER cycle, reading (and possibly modifying)
;		  an indirect address from memory
;   2 - Execute	  is ON for an EXECUTE cycle, executing the requested instruction
;   3 - IR 0
;   4 - IR 1	  (instruction register) latch the opcode of the instruction being
;   5 - IR 2	  executed
;   6 - MD DIR	  indicates whether memory will be re-written with the value in MB
;		  (essentially a read), or a new value will be written from an
;		  internal register (ie, a write). Note that all PDP-8 memory cycles
;		  are read-write, as core requires the data be re-written.
;		  Implemented for ISZ, DCA, and JMS (thus ignoring that read from core
;		  requires a write operation to correct the destructive read operation
;		  of core memory). ISZ, DCA, and JMS do always a write operation, that
;		  is inherent to the type of instruction execution.
;   7 - DATA CONT signals an addition is required during a read-modify-write operation
;		  It is turned on in the (special) case when an instruction does an
;		  indirect access to the lower words at 010-017 in page-0 that exhibit
;		  pre-increment auto-index, and during ISZ execute cycles.
;		  These 2 previous mentioned condition are implemented.
;		  DATA CONT is also turned on during data break cycles to increment
;		  addresses and decrement counters (not implemented),  and maybe (???)
;		  during JMS execute cycles too (not sure about that: not implemented)
;   8 - SW	  SW reflects the state of the SW switch
;   9 - Pause	  is set when an I/O instruction is being executed. It is reset
;		  automatically at the end of the cycle unless the peripheral requests
;		  another cycle for the I/O (not yet implemented ...?)
;  10 - BRK PROG  BRK (in) PROG indicates a data break cycle (DMA) has been requested
;		  (not yet implemented, will it ever be implemented ...?)
;  11 - BRK	  BRK (cycle) means a requested data break (DMA ACK) is in progress
;		  (not yet implemented, will it ever be implemented ...?)
;
;
;  STATUS (STS)	- indicates an individual function for each bit:
;
;   0 - LNK	  LINK is the link (carry) bit
;   1 - GT	  Greater Than flag
;   2 - INT BUS	  Interrupt Bus, indicates that an I/O device has requested an IRQ
;   3 - NO INT	  No Interrupt Allowed, indicates that no interrupt is allowed just
;		  now, even if interrupts are nominally enabled
;   4 - ION	  Interrupt On, indicates interrupts are enabled
;   5 - UM	  User Mode, indicates that the processor is in user mode (IOT, HLT,
;		  OSR generate a trap) - Time Share option
;   6 - IF0
;   7 - IF1	  Instruction Field, these are the 3 EMA bits for a fetch or a direct
;   8 - IF2	  memory reference
;   9 - DF0
;  10 - DF1	  Data Field, the 3 EMA bits for an indirect memory reference
;  11 - DF2
;
;
;  AC		- indicates bits 0-11 of the accumulator AC at TS1
;  MD		- indicates accessed data just written or rewritten into memory
;  MQ		- indicates contents of MQ (Multiplier Quotient) register at TS1
;  BUS		- indicates bits 0-11 of the (OMNIBUS) I/O data lines. It is the
;		  12-bit value being OR-ed or loaded into AC during programmed I/O,
;		  for example the console I/O. DMA is different and uses MA/MD.
;
; One example: if address 0200 contains the instruction "TAD I 10", the addresses 0200
; and 10 are accessed with IF, and the memory to be added to AC is accessed with DF.
;
;
;  Real Console implementation
; -----------------------------
;
;    LED's : 15 - memory address         switches : 15,  keys: 6
;            12 - data register(s)       selector :  6 positions
;             1 - RUN LED                key      :  1 position "LOCK" (& PWR on/off)
;   ----------------------------------------------------------------------------------
;       output bits  : 28                   input bits  : 28
;       output ports : 4                    input ports : 4
;
;      ASSIGNMENT of OUTPUT PINs           ASSIGNMENT of INPUT PINs 
;
;      outport #0 = address low 8 bits      inport #0 = switch 4 - 11
;      outport #1 = address high 7 bits     inport #1 = switch 0 - 3 and SW
;      outport #2 = data low 8 bits                     SW, HALT, SING-STEP
;      outport #3 = data high 4 bits        inport #2 = 6 bit ISS knob
;                   and RUN LED             inport #3 = ADDR-LOAD, EXTD-ADDR-LOAD
;                                                       CLEAR, CONT, EXAM, DEP
;
;  Connection of the switches and keys to the I/O Board
; ------------------------------------------------------
;  I used switches and keys that have *3* positions. Only in the "up" and "down"
;  position a contact is closed. In the middle position none of the 2 contacts
;  is closed. Using 3-position switches gives the possibility to replicate the
;  position of the switches and keys in the best manner. With 3-position switches
;  the logical "up" position is the physical *middle* position of the lever; the
;  physical up position is not used/connected. See also http://www.pdp-11.nl and
;  click the link [Homebrew 'PDP-11'] in the menu at the left, then the link
;  [spin off] at the right top.
;
;  SW, HALT, SING-STEP switch, SR switches:     ADDR-LOAD, EXTD-ADDR-LOAD, CLEAR,
;  up (active) is "1", down position is "0"     CONT, EXAM, and DEP *keys* :
;                                               up is "0", down (activate) is "1"
;                                               * also for the 6 pos. rotary knob
;                    +5
;                   |
;                  ---                          nc ----O 
;                  | | 2200R                         .---o----------------> input
;                  | |                          +5 ----O         |
;                  ---                                          ---
;     nc ----O      |                                           | | 2200R
;          .---o----------------> input                         | |
;        +---O                                                  ---
;        |                                                       |
;       === gnd                                                 === gnd
;
;
; Wiring from panel to I/O Board   (I like to keep all things together in this file)
;
;  =======================================||=========================================
;  |____________ OUT01 connector _________||_____________ OUT23 connector __________|
;  | pin# color code    pin# color code   || pin# color code     pin# color code    |
;  --------------------+------------------||-------------------+--------------------|
;  | 1   yellow        | 2   grey/pink    || 1   black         | 2   pink           |
;  | 3   blue          | 4   red          || 3   green         | 4   grey/pink      |
;  | 5   white         | 6   white/beige  || 5   purple        | 6   yellow         |
;  | 7   brown         | 8   black        || 7   red           | 8   blue           |
;  | 9   not connected | 10  green        || 9   brown         | 10  white          |
;  | 11  red/blue      | 12  brown/yellow || 11  not connected | 12  white/green    |
;  | 13  grey          | 14  purple       || 13  not connected | 14  grey           |
;  | 15  pink          | 16  brown/green  || 15  not connected | 16  white/beige    |
;  ==================================================================================
;
;
;  =======================================||=========================================
;  |_____________ IN01 connector _________||_____________ IN23 connector ___________|
;  | pin# color code    pin# color code   || pin# color code     pin# color code    |
;  --------------------+------------------||-------------------+--------------------|
;  | 1   brown/yellow  | 2   grey/pink    || 1   grey          | 2   brown/green    |
;  | 3   brown         | 4   white        || 3   pink          | 4   purple         |
;  | 5   purple        | 6   red          || 5   not connected | 6   brown/yellow   |
;  | 7   pink          | 8   green        || 7   not connected | 8   red/blue       |
;  | 9   not connected | 10  grey         || 9   blue          | 10  green          |
;  | 11  brown/green   | 12  yellow       || 11  grey/pink     | 12  white          |
;  | 13  white/beige   | 14  red/blue     || 13  not connected | 14  not connected  |
;  | 15  blue          | 16  white/green  || 15  brown         | 16  yellow         |
;  ==================================================================================
;
; Reset of the Core Board via the console keys.
; To maintain the original console panel, but still have a RESET button available for
; the 6809 processor, the CLEAR and the DEP key are wired to a 74LS00 that is glued
; to the rear side of the backplate. The output of the NAND gate is connected to the
; *correct* pin of the RESET header on the Core.
;
;
; Additional notes on the Real Console implementation
;
; - With a (small) performance penalty the lamps F, D, and E can be added with each
;   simulated instruction. When the 6809 code accesses the jump table to the
;   instruction, the Fetch lamp is turned on. In each instruction F is turned off,
;   and E is turned on; when memory access is done, D is also turned on.
;   When the next instruction is taken from the table F goes on, and D and E are
;   turned off.
;
;   Also the IR0-2 indicators get loaded during the fetch step.  PAUSE may also need
;   to be set during the F cycle of an IOT, but I'd have to check.  That might be
;   able to wait for the E cycle.
;   MD should be the data moving to or from memory, so during F it should be the
;   instruction, during D it should reflect the address being read (and possibly
;   modified), and during E it should be updated to reflect the contents of the
;   memory being referred to (except during IOTs and OPRs).
;
;
;
;  Real Console definitions
; --------------------------
;			                 - 7 -- 6 -- 5 -- 4 -- 3 -- 2 -- 1 -- 0 -
WR8e_0	=	0	; outport #0 :    A7   A6   A5   A4   A3   A2   A1   A0
WR8e_1	=	1	;         #1 :   EMA2 EMA1 EMA0   -   A11  A10  A9   A8
WR8e_2	=	2	;         #2 :    D7   D6   D5   D4   D3   D2   D1   D0
WR8e_3	=	3	;         #3 :     -    -    -   RUN  D11  D10  D9   D8

RD8e_0	=	0	; inport  #0 :    S4   S5   S6   S7   S8   S9   S10  S11
RD8e_1	=	1	;         #1 :   HALT SSTP  SW    -   S0   S1   S2   S3
RD8e_2	=	2	;         #2 :     -    -   STE  STS  AC   MD   MQ   BUS
RD8e_3	=	3	;         #3 :    CNT   -   AL   EAL  CLR  -    EXM  DEP
			;		note: CONT key *must* be bit-7 (performance)


  ; inport #RD8e_1
SW_hlt	=	0b10000000	; input bit 7 = HALT switch
SW_stp	=	0b01000000	;       bit 6 = SING STEP switch
SW_sw	=	0b00100000	;       bit 5 = SW switch
SW_swh	=	0b00001111	;       bit [3-0] : mask Switch Reg [0-3]

  ; inport #RD8e_2
ISSste	=	0b00100000	; input bit 5 = STATE position
ISSsts	=	0b00010000	;       bit 4 = STATUS position
ISSac	=	0b00001000	;       bit 3 = AC position
ISSmd	=	0b00000100	;       bit 2 = MD position
ISSmq	=	0b00000010	;       bit 1 = MQ position
ISSbus	=	0b00000001	;       bit 0 = BUS position

  ; inport #RD8e_3
SW_cnt	=	0b10000000	; input bit 7 = CONT key  (** must be bit 7 **)
SW_al	=	0b00100000	;       bit 5 = ADDR LOAD key
SW_eal	=	0b00010000	;       bit 4 = EXTD ADDR LOAD key
SW_clr	=	0b00001000	;       bit 3 = CLEAR key
SW_exm	=	0b00000010	;       bit 1 = EXAM key
SW_dep	=	0b00000001	;       bit 0 = DEP key

ALLKEYS	=	SW_al+SW_eal+SW_clr+SW_cnt+SW_exm+SW_dep

RUNled	=	0b00010000	; output bit where RUN LED is connected

  ; output #WR8e_3 and WR8e_2 (Indicator Selector Switch == "STATE")
LD_fet	=	0b00001000	; output bit 3 = Fetch
LD_def	=	0b00000100	;            2 = Defer
LD_exe	=	0b00000010	;            1 = Execute
LD_ir0	=	0b00000001	;            0 = Instruction msb
LD_ir1	=	0b10000000	;            7 = Instruction nsb
LD_ir2	=	0b01000000	;            6 = Instruction lsb
LD_mdd	=	0b00100000	;            5 = MD DIR
LD_dac	=	0b00010000	;            4 = DATA CONT
LD_swt	=	0b00001000	;            3 = SW
LD_pau	=	0b00000100	;            2 = PAUSE
LD_bkp	=	0b00000010	;            1 = BRK PROG
LD_brk	=	0b00000001	;            0 = BRK

  ; output #WR8e_3 and WR8e_2 (Indicator Selector Switch == "STATUS")
LD_lnk	=	0b00001000	; output bit 3 = Link
LD_gtf	=	0b00000100	;            2 = Greater Than flag
LD_irb	=	0b00000010	;            1 = Interrupt Bus
LD_nia	=	0b00000001	;            0 = No Interrupt Allowed
LD_ion	=	0b10000000	;            7 = Interrupt On
LD_umd	=	0b01000000	;            6 = User Mode
LD_if0	=	0b00100000	;            5 = Instruction Field 0
LD_if1	=	0b00010000	;            4 = Instruction Field 1
LD_if2	=	0b00001000	;            3 = Instruction Field 2
LD_df0	=	0b00000100	;            2 = Data Field 0
LD_df1	=	0b00000010	;            1 = Data Field 1
LD_df2	=	0b00000001	;            0 = Data Field 2


   ; parallel ports (MC6821, PIA)

PIADA	=	0x00E0		; PIA data A register
PIACA	=	PIADA+1		;     control A register
PIADB	=	PIADA+2		;     data B register
PIACB	=	PIADA+3		;     control B register

   ; PIA [port A] ConTroL register bitmask
   ; - during execution reconfigured to input/output port
   ; - CA2=output, IRQ on CA1 enabled/disabled

PACTLI	=	0b00111100	; CA2="1", buffer LS245 -> input
PACTLJ	=	0b00111101	; as PACTLI, but CA1 irq enabled
PACTLO	=	0b00110100	; CA2="0", buffer LS245 -> output
PACTLQ	=	0b00110101	; as PACTLO, but CA1 irq enabled
DIRBIT	=	0b00000100	; access data-in/out or data-dir
CA2BIT	=	0b00001000	; CA2 bit: 245-buffer ctrl in/out
				; 0 :: output, 1 :: input
IOADDR	=	0b00111111	; bit [5:0] - board & port select address bits
RDWR  	=	0b01000000	; bit [6] - RD/WR direction:  0::read   // 1::write
DIS3S 	=	0b10000000	; bit [7] - LS138's control:  0::enable // 1::disable


   ; PIA [port B] ConTroL register bitmask
   ; - always configured as output port
   ; - CB2=output set to "0" (buffer LS245 -> output)
   ; - IRQ on CB1 disabled/enabled

PBCTLO	=	0b00110100	; CB1 irq disabled
PBCTLQ	=	0b00110101	; CB1 irq enabled


  ; Input ports and output ports states
  ; - NOTE: keep these location CONSECUTIVE
  ;        (processed as an array thru indexing with offsets)

  ; memory allocation starts at <RCRAM>

  ; * Interrupt routine used storage (for input switch/toggle processing )
  ; * Accessed through an index, base address == <RCRAM>.
  ; * Number between brackets [] indicates the number of bytes required

CONIHI	=	3		; highest input port number used

SWTDBC	=	0		; [CONIHI] debounce timers for each port
PREVIN	=	SWTDBC+CONIHI+1	; [CONIHI] intermediary storage locations
INPORT	=	PREVIN+CONIHI+1	; [CONIHI] input ports (for application)
LSTIRQ	=	INPORT+CONIHI	; last interrupt routine index accessed location
LSTIRX	=	LSTIRQ+1	; 1 byte safety margin

BOUNCE	=	8		; # ticks used as switch/toggle debounce delay
TFRDLY	=	BOUNCE+5	; response delay for toggle release/activate






; This program emulates a pdp8/e on the 6809 Core board computer.  It is based
; on the PDP8.ASM program, with the addition of the KM8-E Memory Extension and
; the Time Share module.
;
;
;     PROBLEMS TO SOLVE  /  STUFF TO ADD
;    ------------------------------------
;
; 1. Implement the connection to the RealConsole *native* !!
; 2. The BIN loader in this implementation reads a leader, the data part and
;    then a trailer. Papertapes exist that have multiple parts, so the trailer
;    is also the leader for the next part. The "real" bin loader used flow
;    control to stop the reader when it was halted.
;    Then you could read the checksum and press "CONTinue".
; 3. Implement a good solution for flow control. RTS/CTS is possible, but not
;    from the ACIA on the Core Board. The use of one in- and output pin of the
;    I/O Board's ports produces a workable situation ...(?)
; x.
;
;
;
; The PDP-8 is a 12-bit computer with 4096 words of memory. It has only eight
; instructions, a 12-bit program counter, the 12-bit accumulator AC, and a
; single-bit register, the link (L) bit, prefixed to AC as <L,AC>.
; All instructions have the following 12-bit format:
;
;  11  10  9   8   7   6   5   4   3   2   1   0
;  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ 
; |   |   |   |   |   |   |   |   |   |   |   |   |  i - indirect bit [0=direct,
; |_ _|__ |__ |__ |__ |___|___|__ |__ |__ |__ |__ |                    1=indirect]
; :           :   :   :                           :  z - page bit [0=page zero, 
; :  opcode   : i : z :           address         :                1=current page]
;                                                    address - 12-bit word in page
;
; Note. Absolute memory locations 8 through 15 are incremented prior to use when
;       used as indirect addresses. These locations are called auto-index registers.
;
; The 8 opcodes (instructions) are the following:
;
;  000 - AND - logical AND of AC with memory
;  001 - TAD - 2-complement add of memory to <L,AC> (13-bit value)
;  010 - ISZ - increment memory and skip next instruction if result is 0
;  011 - DCA - deposit AC in memory and clear AC
;  100 - JMS - jump to subroutine. Current PC is stored in subroutine start address.
;  101 - JMP - jump unconditional
;  110 - IOT - input/output transfer
;  111 - OPR - microcoded operations, 3 groups
;
;
; Emulator data:
;
; (1) Emulated program space is 2 (8-bit) bytes for each PDP-8 (12-bit) location.
; (2) Non existent fields are read as 0.
; (3) Interrupt inhibit (removed by a JMP or JMS) is indicated by the first
;     bit in the interrupt flags being set.  Instruction delay (due to ION) is
;     handled by changing a pointer. If the pointer indicates that the interrupts
;     are already enabled it is not changed.  It is assumed that ION delay and
;     JMP/JMS inhibit are not counted in checking if the interrupt system is on.
;     See the JMP code for further explanation.
;
; Memory data format:
;
;   ooooCCCIZAAAAAAA
;
;	o = Must be 0 (only 12 bits used)
;	C = Opcode
;	I = Indirect address if 1
;	Z = Current page if 1
;	A = Page address
;
; Memory Extension and Time Share:
;
;    There are three register sets: Saved, Working and Buffer.
;    Each set has both Field numbers and pointers.  The User field is really
;    a flag for User/Executive mode.  If the bit is 0 it is in the Executive
;    (standard) mode, else it is in the User mode in which all LAS, HLT and IOT
;    instructions are bypassed and the Time Share interrupt set. The registers
;    include the memory pointers and the field number bytes, as follows:
;
;	  U= Time Share flag (0= Executive, 1= User).
;	III= Instruction field number.
;	DDD= Data field number.
;
;	__________________________________________________________
;	| Function  || RAM Registers                               |
;	|           ||_____________________________________________|
;	|           || I field | D field | Field bytes and bit map |
;	|           || pointer | pointer |_________________________|
;	|           ||         |         | U and I    | D          |
;	|           ||         |         | (U0000III) | (00000DDD) |
;	|===========||=========|=========|============|============|
;	|   Working ||  WrkIp  |  WrkDp  |     WorkUI |      WorkD |
;	|-----------||---------|---------|------------|------------|
;	| Interrupt ||  SavIp  |  SavDp  |     SaveUI |      SaveD |
;	|-----------||---------|---------|------------|------------|
;	|   J Delay ||  BufIp  |         |     BuffUI |            |
;	|___________||_________|_________|____________|____________|
;
;
; Modifications:
;
; (1) 4-8-97	F. Wilson: Fixed small errors in the Extended Memory and
;			 Time Share module.
;		Is not tested yet, and will need update of the test compiler.
; (2) 4-15-97	F. Wilson: Added checks for breakpoints and a monitor.
;			 Single stepping is the starting default.
; (3) 5-2-97	F. Wilson: Put Instruction Field into GTF.
; (4) 5-6-97	F. Wilson: Removed single step routine and incorporated into
;			 a primitive monitor. Checked emulator speed with
;			 the LED display on and it drops by over 40%.
; (5) 5-12-97	F. Wilson: Removed indirects from field pointer access in field
;			 change IOT's.
; (6) 6-11-97	F. Wilson: Changed vector lists to end in 0.
; (7) 9-3-97	F. Wilson: Put ^C exit into the binary loader.
;		Made location 200 in Field 0 the default program start.
; (8) 6-14-04	H. Gooijen: Adapted the 6809 code to be used with the Core Board,
;			 see for more info http://www.pdp-11.nl [homebrew PDP-11].
;		     The code can now be put in EPROM. The self-modifying code is
;		     removed and a 'decent' implementation has replaced it.
;		     Changed the source code to be compatible with the MACRO-11
;		     syntax, as the code is assembled with the ASxxxx assembler of
;		     A. Baldwin.  The original implementation used 2 bits of a PIA
;		     to create a serial interface. As the Core Board has a real
;		     ACIA, the PIA code is removed. With this the option to change
;		     the baudrate is gone too, though ...
;
;  Speed improvements
; --------------------
; 1) moved the GetAdr routine so that some often used pdp8 instructions can reach
;    that routine through "bsr" instead of "jsr".
; 2) to update the state flags, the contents of an accumulator must be saved (in
;    most instances). That was done through the stack (pshs/puls).  The psh/pul
;    instruction both need 5 cycles. When sta/lda is used accessing a direct page
;    address, both these instructions need only 4 cycles.  Thus this sequence is
;    20% faster. The memory location *used solely* for this purpose is "savacc".
; 3) The "ISZsub" routine uses at the end the sequence "std  ,u" followed by a
;    "cmpd  #0" and a "bne" instruction. Since the "std" instruction already sets
;    the Z condition code, the cmpd instruction can be removed.
; 5) The sequence clra/clrb uses 2 bytes, wehereas ldd #0 gives the same result
;    but uses one byte more. The "ldd #0" is preferred as clra/clrb takes 4 cycles
;    and the ldd #0 takes 3 cycles (one less).
; 6) ... update the Core Board ...
;    change the 1 MHz parts (6809, 6821, 6850) to the 2 MHz version ("B")
;    and change the Xtal oscillator from 4 MHz to 8 MHz.
;    Check the access speed of the EPROM and the RAM chips!
;    This makes *everything* twice as fast!!
; x) ...
;
;
;
; To do:
; (1) Add "paper tape" load and punch.  Perhaps can have a buffer for main
;     program download to decrease the binary loader delay and then re-use the.
;     Use up and down loading from the PC, with time delay to indicate end of data.
; (2) On HLT condition keep the console alive instead of busy wait, loop for reset.


	.module	pdp8e
	.radix	d
	.area	WORKPG	(PAG)
	.setdp	0,WORKPG

; ================================================================================

BINldr_txt	=	0  ; 1 :: print BIN loader text identifying received data
			   ; 0 :: do not print any text

RealCons	=	1  ; 1 :: include the RealConsole!
			   ; 0 :: do not include it!

; ================================================================================


    ; 6809 monitor I/O routines & hardware address locations

docrlf	=	0xC02D	; output CR/LF to console
ino	=	0xC00C	; input character (8 bits)
inch	=	0xC009	; get keyboard entry in [A], no echo
OUTB2O	=	0xC021	; print data in [A] as 2 hexadecimal digits
outch	=	0xC024	; print character in [A] - not destroyed
OUTABT	=	0xC030	; print character in [A], [B] times repeated
OUT2CH	=	0xC027	; output 2 chars [A] then [B]
OUT4HS	=	0xC039	; print [X] as 4 hex followed by one space

.firq	=	0x21F6	; FIRQ vector location in monitor RAM
.irq	=	0x21F8	; IRQ vector
.swi	=	0x21FA	; SWI vector
.nmi	=	0x21FC	; NMI vector

ACIACS	=	0x00F0		; ACIA control & status
ACIADA	=	ACIACS+1	; ACIA data I/O
ACIANI	=	0b00010001	; /16, 8N2, receiver IRQ disabled
ACIAI	=	0b10010001	; /16, 8N2, receiver IRQ enabled
RTSBIT	=	0b01000000	; RTS off / on bit  


    ; ASCII constants

CR	=	0x0D		; Carriage return
LF	=	0x0A		; Line feed
CtrlC	=	0x03		; Control C
Space	=	0x20		; Space
BS	=	0x08		; Backspace
XON	=	0x11		; Data link "on"  prompt (DC1 or ^Q)
XOFF	=	0x13		; Data link "off" prompt (DC3 or ^S)

    ; program equates

Opmask	=	0b00001110	; Upper byte opcode mask
Wmask	=	0b00001111	; Upper byte memory data mask
Wsize	=	0o7777		; Memory data mask
Fsize	=	0o10000*2	; 4096 words/memory location
Offset	=	0b01111111	; Page offset bits
LoPage	=	0b10000000	; Current page (low byte)
Lmask	=	0b00010000	; Isolate link
ACmask	=	0b00001111	; Drop link
IOTmask	=	0b01111110	; Isolate IOT address
LtoC	=	0b11110000	; Add to link (moves to carry)
IOsetup	=	8		; IO device setup ("impossible" IOT)
IOreset	=	9		; Clear/CAF

Ubit	=	0b10000000	; Field number bit masks
Fbits	=	0b00000111
Dbits	=	0b00000111
Jdelay	=	0b10000000	; JMP/JMS interrupt inhibit
UserI	=	0b01000000	; User Interrupt bit
SetIO	=	0b0010000000000000	; Start bit for interrupt flags

TRUE	=	1
FALSE	=	0


    ; Simulated PDP-8 memory.
    ; Maximum theoretical is 64 Kbytes, which is the full 6809 addressing space.
    ; Originate RAM at easily handled boundary.

RAMst	=	0x2300		; start of free RAM on Core Board ****
Field0	=	RAMst		; $2300-$42FF
Field1	=	Field0+Fsize	; $4300-$62FF
Field2	=	Field1+Fsize	; $6300-$82FF
Field3	=	Field2+Fsize	; $8300-$A2FF
;Field4	=	Field3+Fsize
;Field5	=	Field4+Fsize
;Field6	=	Field5+Fsize



    ; storage in WORKPG

PC:	.rmb	2	; Program counter
AC:	.rmb	2	; Accumulator
MQ:	.rmb	2	; Multiplier quotient
Link:	.rmb	1
MD:	.rmb	2	; Present instruction
SR:	.rmb	2	; Front panel switch register
Temp2:	.rmb	2
Temp1:	.rmb	1
IOflgs:	.rmb	2	; Input output device interrupt requests
Uflag:	.rmb	2	; User interrupt bit
FLGset:	.rmb	2	; Temporary for device flag setup
Dcount:	.rmb	1	; Temporary for device #
IOctrl:	.rmb	1	; Temporary device IOT
Idelay:	.rmb	1	; ION delay count

    ; Following locations are used by the Memory Extension and Time Share.
    ; Keeping the User flag in the MSB allows a simple sign check.
    ; Have WorkD follow WorkUI and SaveD follow SaveUI so can transfer as a word.

WrkIp:	.rmb	2	; Working I field (also use X for pointer)
WrkDp:	.rmb	2
WorkUI:	.rmb	1
WorkD:	.rmb	1
SavIp:	.rmb	2	; saved during interrupt processing
SavDp:	.rmb	2
SaveUI:	.rmb	1
SaveD:	.rmb	1
BufIp:	.rmb	2	; loaded into working U and F by JMP or JMS
BuffUI:	.rmb	1


    ; flags for "display" and vectored jumps to emulation function
    ; - added to get rid of self-modifying code!

emufnc:	.rmb	2	; vector address is: Doinst, Chkdly, or Chkint

    ; memory word to quick access the 'lamp values' for the different
    ; settings of the Indicator Selector Switch (ISS).  The words are
    ; kept "up-to-date" during the execution of the simulation at many
    ; locations to facilitate a reasonable correct RealConsole display.

savacc:	.rmb	1	; save location for updating the following data
rc_mad:	.rmb	2	; EMA + Memory Address
rc_ste:	.rmb	2	; ISS -> STATE
rc_bus:	.rmb	2	; ISS -> BUS


    ; local storage used in the subroutines

binptr:	.rmb	2	; [binary loader] absolute load location
bindat:	.rmb	2	; [binary loader] previous memory data
binsum:	.rmb	2	; [binary loader] checksum
binsv0:	.rmb	2	; [binary loader] last input data (low-byte)
binsv1:	.rmb	2	; [binary loader] present input data

Ibit:	.rmb	2	; [TTY] Interrupt flag bit
Icom:	.rmb	2	; [TTY] complement of flag bit
Istat:	.rmb	1	; [TTY] Interrupt enable (FALSE == not enabled)
			; [TTY] keep following two together:
TTIflg:	.rmb	1	; [TTY] character received flag (TRUE == received)
TTOflg:	.rmb	1	; [TTY] output character done flag (TRUE == done)
TTIdat:	.rmb	1	; [TTY] input character storage

issts2:	.rmb	2	; [ISSSTS] scratch word

_bin3:	.rmb	1	; [binwrd]
_bin4:	.rmb	1	; [binwrd]

_stp2:	.rmb	1	; formatted copy of Link
_stp3:	.rmb	1	; formatted User mode
_stp4:	.rmb	1	; true if interrupt system is on
_stp5:	.rmb	1	; true if interrupt requests blocked
_stp6:	.rmb	1	; true if have interrupt requests

_octp2:	.rmb	1	; [PrtOct] (print octal number) digit counter
_octp3:	.rmb	1	; [PrtOct] (print octal number) character counter
_octp4:	.rmb	1	; [PrtOct] (print octal number) data to print
_octp5:	.rmb	1	; [PrtOct] (print octal number)



	.if	RealCons
piadir:	.rmb	1	    ; indicator: PIA [port A] current data direction
FIRQsv:	.rmb	1	    ; save location, used *ONLY* local in FIRQ routine
rtssts:	.rmb	1	    ; RTS* pin status of ACIA
timcnt:	.rmb	1	    ; time-out interval setpoint value
timer:	.rmb	1	    ; running time-out interval value
RCRAM:	.rmb	LSTIRX	    ; assign index-referenced storage
procin:	.rmb	1	    ; current processing input port #

INBASE	=	RCRAM+INPORT  ; for easy reference

dlyexm:	.rmb	1	    ; EXAMINE toggle response control
dlydep:	.rmb	1	    ; DEPOSIT toggle response control
dlyclr:	.rmb	1	    ; CLEAR toggle response control
dlyal:	.rmb	1	    ; ADDRESS LOAD toggle response control
dlyeal:	.rmb	1	    ; EXTENDED ADDRESS LOAD toggle response control

conact:	.rmb	1	    ; console action response
ACTrun	=	0b10000000  ; action = run
ACTstp	=	0b00000001  ;        = step instruction
ACTcyc	=	0b00000010  ;        = cycle instruction

	.else

CCsave:	.rmb	1	    ; save interrupt mask
readpc:	.rmb	2	    ; entry of 4-digit octal number (used to get PC)
MONflg:	.rmb	1	    ; Monitor call flag

    ; Monitor flags

Console	=	0b10000000	; Call monitor (from keyboard ^S)
Break0	=	0b01000000	; Break point #0
Break1	=	0b00100000	; Break point #1
Break2	=	0b00010000	; Break point #2
Halt	=	0b00001000	; Halt instruction or HALT switch active
I_path	=	0b00000100	; Interrupt service
SnglStp	=	0b00000010	; Single stepping
Breaks	=	Break0+Break1+Break2

	.endif


RDRin:	.rmb	2	    ; store into buffer pointer
RDRout:	.rmb	2	    ; retrieve from buffer pointer
RDRblo	=	0xBC00	    ; buffer start address *** hard coded ***
RDRbhi	=	0xC000	    ; buffer end (not inclusive)

	.if	BINldr_txt
RDRddd:	.rmb	1	    ; count printed "d" for data block
RDRltr:	.rmb	1	    ; count printed "." for leader/trailer block
RDRcnt:	.rmb	1	    ; count # printed characters on one line
	.endif


	.area	PDP8E	(ABS,OVR)

	; Start the pdp8/e emulator

X_PDP8:	ldd	#WORKPG	    ; set all var's in "zero-page"
	tfr	a,dp
	ldx	#Sysmsg	    ; issue identification message
	jsr	pdatao
	bsr	RAMclr	    ; clear (existing) memory
	.if	RealCons
	jsr	initRC	    ; initialize RealCons hardware
	.else
	lda	#Console    ; call Monitor on start
	sta	*MONflg
	.endif

	jsr	PDPld	    ; load at least one binary file

	.if	RealCons
	ldd	#0o200	    ; set a common default start PC
	.else
getpc1:	ldx	#askpc
	jsr	pdatao	    ; ask for start PC
	ldb	#4	    ; get 4 octal digits max
	jsr	getoct
	bvs	getpc1	    ; entry aborted - try again
	bcc	getpc2	    ; good entry, new PC accepted
	ldd	#0o200	    ; <CR> - set default start PC (0200)
getpc2:
	.endif

	std	*PC	    ; ** overrule by Console possible
	lda	#IOsetup    ; set up IO devices (KL8-E turns on IRQ)
	bsr	ClrSys
	clra		    ; IF0-1-2 and DF0-1-2 set to Field zero
	sta	*WorkUI	    ; save
	sta	*WorkD
	sta	*BuffUI	    ; same as Extended Address Load on panel
	lsla		    ; get field pointer
	ldu	#GetFld
	ldx	a,u
	beq	Flderr	    ; no memory allocated
	stx	*WrkIp
	stx	*WrkDp
	stx	*BufIp
	clra		    ; set the Save Field registers to field 0
	sta	*SaveUI	    ; NOTE: Real PDP-8 may have random value
	sta	*SaveD
	ldd	0,u	    ; (keep X set to the Instruction field)
	std	*SavIp
	std	*SavDp

    ; initialize display
    ; the routine "initRC" enabled the FIRQ that processes the switches/keys.
    ; wait a little to obtain stable switch/key states in RCRAM memory array.

	.if	RealCons
	ldb	#3	; delay 300 mS
inidly:	ldy	#14285	; delay for 100 mSec (at 1 MHz CPU clock)
1$:	leay	-1,y
	bne	1$
	decb
	bne	inidly
;
	.endif

    ; Switch register usually at start address,
    ; as standard startup sequence is Address Load, Clear and Continue.

	ldd	*PC	; set switch register to start address
	std	*SR
	jmp	Emulat	; start

    ; Memory not allocated for the specified field

Flderr:	jsr	Errpnt
	.fdb	NoFLD



;  * RAMclr - clear existing RAM fields

;    The memory locations are filled with HLT (7402) or zeroes (0000).
;    The available memory is indicated by the GetFld table.

RAMclr:	lda	#8	; max is 8 blocks of 8kbyte RAM (4k word)
	pshs	a	; push block counter on stack
	ldy	#0o7402	; fill with HLT (7402)
; -->	ldy	#0	; fill with zero (0000)
	ldu	#GetFld	; base of available memory table
ramnxF:	ldx	,u++	; get base address of 8kbyte block
	beq	ramnoF	; non-existent -
	ldd	#Fsize	; set up size (words) of one block
ramclF:	sty	,x++	; store HLT or zero in entire block
	subd	#2
	bne	ramclF	; until block filled/erased
ramnoF:	dec	0,s	; all blocks available checked?
	bne	ramnxF	; n -
	puls	a,pc


    ; Field location array.
    ; If location == 0 the field does not exist. Field0 must always exist.

GetFld:	.fdb	Field0	; -> must exist
	.fdb	Field1	; -> minimum RAM requirement for OS8
	.fdb	Field2
	.fdb	Field3
	.fdb	0	; Field 4 (system RAM)
	.fdb	0	; Field 5 (system RAM)
	.fdb	0	; EPROM - monitor space
	.fdb	0	; EPROM - monitor space




    ; Panel "start" or CAF. Clears the AC, MQ and link. Does not affect the data and
    ; instruction field settings.  Sets the Time Share to executive mode (U= 0).
    ; Peripheral flags cleared, but interrupt requests (usually) enabled.
    ; Interrupt system is off and ION and JMP/JMS inhibits cleared.
    ; On exit D == 0.

Clear:	lda	#IOreset	; reset flags. Enable interrupt requests
ClrSys:	clr	*IOflgs
	clr	*IOflgs+1
	jsr	IOset		; enter with A = IO setup flag
	bne	clrerr
	ldd	#Doinst
	std	*emufnc		; set vector to processing routine
	clra
	clrb
	std	*AC		; clear "registers"
	std	*MQ
	std	*MD
	std	*rc_mad
	std	*rc_ste
	std	*rc_bus
	sta	*Link
	rts
clrerr:	jsr	Errpnt
	.fdb	Clrmsg



; The ISZ and other skip instructions conditionally skip the next
; instruction in sequence.  The ISZ is commonly used to increment a loop
; counter and skip if done, and it is also used as an general increment
; instruction, either followed by a no-op or in contexts where it is known
; that the result will never be zero.

    ; ISZ increments the data at the address and skips the next instruction
    ; if it becomes 0. AC and Link not affected.
    ; Routine moved here so the code can fall through into the PC skip.

ISZsub:	jsr	GetAdr	; clears FETCH flag in STATE
	beq	NextI
	lslb		; translate to 16-bit address
	rola
	leau	d,u
	ldd	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	orb	#LD_dac+LD_mdd	; set DATA-CONT and MD-DIR flag in STATE
	std	*rc_ste
	ldd	,u	; get data from address
	addd	#1	; +1
	anda	#Wmask	; prevent 12-bit overflow
	std	,u
; ---	cmpd	#0	; data result == 0000 ?  [ccZ set by previous std]
	bne	NextI	; n -

;   skip next instruction

PCskip:	ldd	*PC
_skip:	addd	#2
	bra	_savpc

;   next instruction routine

NextI:
;	lda	*PC
;	ora	*AC	     ; ***  TEMPORARY CHECK ROUTINE  ***
;	ora	*MQ	     ; Check data at last step to see
;	ora	*MD	     ; if have any extra bits set
;	bita	#0b11110000  ; all four high bits must be 0
;	bne	b4err
;	lda	*Link
;	bita	#Lmask^0xFF
;	beq	GetI
;b4err: jsr	Errpnt
;	.fdb	Bitmsg


GetI:	ldd	*PC	     ; get the PC
			     ; ***  END CHECK  ***

_nxtni:	addd	#1	     ; update to next
_savpc:	anda	#Wmask	     ; prevent 12-bit overflow
	std	*PC


    ; Following is a jump to one of the following routines,
    ; depending on the interrupt status, set-up in <emufnc>.
    ;  - Doinst  - Chkdly  - Chkint

Emulat:	jmp	[emufnc]     ; jump target can change

;  ________________________________________________________________________________
;  ################################################################################
;  ================================================================================



;  ##===========================================##
;  ##						##
;  ##    Emulation "target" routine:  Chkdly    ##
;  ##						##
;  ##===========================================##

    ; Process ION delay count. When count is 0 set Emulat to
    ; go to interrupt check and then fall through into check.

Chkdly:	dec	*Idelay	; ION delay finished?
	bne	Doinst
	ldd	#Chkint	; change Emulat object
	std	*emufnc	; set vector to processing routine



;  ##===========================================##
;  ##						##
;  ##    Emulation "target" routine:  Chkint    ##
;  ##						##
;  ##===========================================##

    ; Process interrupts.
    ; The interrupts have to be enabled and the ION delay (1 istruction)
    ; and JMP/JMS delay (due to I field load) finished.

Chkint:	ldd	*IOflgs	; get device flags
	ble	Doinst	; check for request (exit if none or J delay)
	ldd	#Doinst	; turn off interrupts
	std	*emufnc	; set vector to processing routine
	ldd	*WorkUI	; save U, I and D
	std	*SaveUI
	ldd	*WrkDp
	std	*SavDp
	stx	*SavIp
	ldx	GetFld	; set F and D to field 0 & U to exec
	stx	*WrkIp
	stx	*WrkDp
	ldd	*PC	; save return address at location 0
	std	,x
	ldd	#0
	std	*WorkUI
	incb		; do next instruction
	std	*PC
	lslb
	bra	_getI



;  ##===========================================##
;  ##						##
;  ##    Emulation "target" routine:  Doinst    ##
;  ##						##
;  ##===========================================##

    ; go to next instruction

Doinst:	ldd	*PC	; restore PC data
	lslb		; multiply by 2 for relative location
			; (16-bit word :: one PDP8 instruction)
	rola
_getI:	ldd	d,x	; get instruction
	std	*MD

    ; set STATE for new instruction
    ;
    ;   ==> copy IR0-IR1-IR2 from MD
    ;   ==> set the FETCH flag bit
    ;   ==> clear DEFER, EXECUTE, MD-DIR, DATA-CONT
    ;   ==> SW flag bit reflects position of the SW console switch
    ;   ==> clear PAUSE, BRK-PROG, and BRK (not implemented, always 0)

				; [A] = ....012i             (from MD)
	lsra			; [A] = .....012  [Cc] = ind
	lsra			;       ......01  [Cc] = IR2
	rora			;       2......0  [Cc] = IR1
	rora			;       12......  [Cc] = IR0

	ldb	#LD_fet		; FETCH always set, DEFER & EXECUTE clear
	bcc	_ste1		; test IR0 in [Cc], clear -
	orb	#LD_ir0		; IR0 bit is set, set -
_ste1:	stb	*rc_ste

	.if	RealCons
	ldb	*INBASE+RD8e_1	; get HALT, SSTEP, and SW
	bitb	#SW_sw		; SW switch active ?
	beq	_ste2		; - clear
	ora	#LD_swt		; - set: add to IR1-IR2 bits
	.endif
_ste2:	sta	*rc_ste+1	; DATA-CONT, MD-DIR always clear
				; PAUSE, BRK-PROG, BRK always clear


    ; next instruction ready for decode.
    ; State checks; implement flags to execute a check or bypass the check
    ; !!! NOTE : preserve [X] (= instruction field location) !!!


;     _______________________________________________________
;  ###							     ###
;  ###  process the switches and the toggles of the console  ###
;  ###_______________________________________________________###


	.if	RealCons

   ; [conact] flags   -- mutual exclusive --
   ;  ___ ___ ___ ___ ___ ___ ___ ___
   ; | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
   ; |___|___|___|___|___|___|___|___|
   ;
   ;   * running                   * cycle step request
   ;                           * instruction step request
   ;
   ; If none of the bits is set none of the operation must be executed.
   ; This causes the code to loop checking the console, which will set
   ; one of the [conact] flags eventually.

	lda	*conact		; console action response needed ?
;	bita	#ACTrun		; "run" ?  (performance: ACTrun at bit 7)
;	bne	Stpxit
	bmi	Stpxit
	beq	Emulat		; no entry, no active state - await console action

	clr	*conact		; remove console step or cycle *request*
; NYI	bita	#ACTstp		; test for "cycle" or "step" instruction
;				; for now always "step" (cycle Not Yet Implemented)

Stpxit:	lda	*MD		; get instruction opcode bits
	anda	#Opmask		; isolate them
	ldu	#Ops		; already "opcode*2" for word-sized table index
	jmp	[a,u]		; execute instruction --->

	.else

	tst	*MONflg	    ; Monitor called if MONflg != 0
	bne	PDPmon
Stpxit:	lda	*MD	    ; get instruction opcode bits
	anda	#Opmask	    ; isolate them
	ldu	#Ops	    ; already "opcode*2" for word-sized table index
	jmp	[a,u]	    ; execute instruction --->


    ; Enter monitor if:
	; - HLT, keyboard break (^S)
	; - or single stepping
	; - breakpoint(s) became true (not implemented, future use?)

    ; NOTES:
	; (1) If HLT: Was a direct jump and special processing is needed.
	; (2) On entry, may want to drop single stepping.

    ; Process any breakpoints and then give a menu.
    ; Include all the usual stuff, such as:
	; Print memory
	; Set various breaks:
	;  - Memory location / Memory data / Instruction cycles
	; Load switch register (for OSR) / Single Step
	; Load address / Load extended address / Clear / Continue

    ; Start Monitor:
	; Presently only monitor startup (from a keyboard ^S call) and
	; single step flags are serviced.
	; In order to allow TTY output during stepping the interrupts
	; are off in this mode.

    ; Monitor flags:
    ;  ___ ___ ___ ___ ___ ___ ___ ___
    ; | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
    ; |___|___|___|___|___|___|___|___|            label:
    ;                                             ---------------------
    ;                               * not used
    ;                           * single step      SnglStp
    ;                       * interrupt service    I_path
    ;                   * halt instruction         Halt
    ;               * breakpoint #2                Break2 (future use?)
    ;           * breakpoint #1                    Break1 (future use?)
    ;       * breakpoint #0                        Break0 (future use?)
    ;   * call monitor (from keyboard Ctrl-S)      Console


PDPmon:	lda	*MONflg	        ; check for keyboard call
	bita	#Console
	beq	Single	        ; branch if not (default)
	anda	#Console^0xFF   ; remove keyboard call
	ora	#SnglStp	; assume want single stepping
	sta	*MONflg
	tfr	CC,a		; save condition codes (IRQ flags)
	sta	*CCsave
	orcc	#0x10		; disable IRQ interrupt
	ldx	#SSmsg
	jsr	pdatao		; print options & fall through to "Single"


    ; Print system data and then wait for
    ;   - Space to continue
    ;   - K to enter keyboard character
    ;   - CR to exit stepping (start run)
    ;   - 1..6 for register display in ASCII artwork

Single:	jsr	Pntreg	; show system state
_sschk:	jsr	inch	; await keyboard command entry
	cmpa	#CR
	bne	Chkspc
	lda	*MONflg	; drop single stepping
	anda	#SnglStp^0xFF
	sta	*MONflg
	jsr	docrlf	; clear up screen a bit
	lda	*CCsave	; restore interrupts
	tfr	a,CC
	bra	_ssxit

Chkspc:	cmpa	#Space
	beq	_ssxit	; next step
	cmpa	#'?
	bne	_sskbd
	ldx	#SSmsg
	jsr	pdatao	; print command options
	bra	Single	; and show register data again

_sskbd:	cmpa	#'1
	blo	_sskbk	; not numeric
	cmpa	#'6
	bhi	_sskbk	; or out of numeric range (1-6)
	suba	#'0+1	; correct base for table
	ldx	#ISScmd
	lda	a,x	; get selection hardware switch value
	jsr	ShwReg	; display selected register in ACSII art
	bra	_sschk

_sskbk:	anda	#0b01011111 ; drop lower case bit
	cmpa	#'K	; keyboard input?
	bne	_sschk	; n - ignore all other entries
	ldx	#Kbdmsg
	jsr	pdatao	; ask for keyboard input
	jsr	inch	; get it
	jsr	outch	; echo entry
	jsr	ttisub	; install it
	jsr	docrlf	; new line and continue

_ssxit:				; --> exit the monitor
	lda	*MONflg		; remove remaining breaks
	anda	#Break0|Break1|Break2|I_path^0xFF
	sta	*MONflg

	ldx	*WrkIp		; restore CPU registers
	jmp	Stpxit		; continue emulation


   ; translation table: command number entry to Indicator Selector Switch value
   ;	"<1>:STATE  <2>:STATUS  <3>:AC  <4>:MD  <5>:MQ  <6>:BUS"

ISScmd:	.fcb	ISSste,ISSsts,ISSac,ISSmd,ISSmq,ISSbus

	.endif


   ; pdp8/e opcodes routine start address table

Ops:	.fdb	ANDsub
	.fdb	TADsub
	.fdb	ISZsub
	.fdb	DCAsub
	.fdb	JMSsub
	.fdb	JMPsub
	.fdb	IOTsub
	.fdb	OPRsub


    ; Illegal instruction trap

Illins:	jsr	Errpnt
	.fdb	Insterr




    ; Access all IO devices. Enter with the IOP code in register A.
    ; Only codes >= 8 used.
    ; If return with = status all existing devices were able to comply.
    ; On an error the routine takes an immediate exit and returns the
    ; exit code in A.

IOset:	sta	*IOctrl
	ldu	#Device
	ldd	#SetIO	; set up initial flag
	std	*FLGset
	lda	#64	; device count
	sta	*Dcount
ioloop:	ldd	,u++	; check for device exists
	cmpd	#Nodev
	beq	iolxit
	lda	*IOctrl	; get IOT
	pshs	u	; device changes U
	jsr	[-2,u]	; do setup
	puls	u
	bne	seterr
iolxit:	dec	*Dcount	; done?
	bne	ioloop
seterr:	rts		; return == 0 status if no error


    ; Get device flag bit in D.
    ; Used by IO devices during initial setup.
    ; If return 0 are out of flags and device must return != status.
    ; NOTE: FLGset must be initialised with the SetIO value.
    ; Are 15 flags available. The MSB is reserved for interrupt processing
    ; inhibit till JMP/JMS. Next bit is for the Time Share interrupt.

GetFlg:	ldd	*FLGset	; get flag
	pshs	d
	lsra		; update flag
	rorb
	std	*FLGset
	puls	d,pc



    ; AND the AC with the data at the address.
    ; Link not affected.
    ; Indirect will use the data field. Non existent data field read as 0.

ANDsub:	bsr	GetAdr	; clears FETCH flag in STATE
	beq	dcaxit	; check for indirect and no field
_doand:	lslb
	rola
	sta	*savacc
	lda	*rc_ste
	ora	#LD_exe	; set EXECUTE flag in STATE
	sta	*rc_ste
	lda	*savacc
	ldd	d,u
	anda	*AC
	andb	*AC+1
	std	*AC
	jmp	NextI


    ; TAD similar to AND, except that the Link is complemented by any overflow

TADsub:	bsr	GetAdr	; clears FETCH flag in STATE
	lbeq	NextI
	lslb
	rola
	sta	*savacc
	lda	*rc_ste
	ora	#LD_exe	; set EXECUTE flag in STATE
	sta	*rc_ste
	lda	*savacc
	ldd	d,u
	ora	*Link
	addd	*AC
	sta	*savacc	; temporary save: preserve Link bit
	anda	#ACmask
	std	*AC
	lda	*savacc	; get Link bit back
	anda	#Lmask
	sta	*Link
	jmp	NextI


    ; DCA similar to AND. AC stored at the address and then cleared.
    ; Link is not affected.

DCAsub:	bsr	GetAdr	; clears FETCH flag in STATE
	beq	dcaxit
	lslb
	rola
	leau	d,u
	ldd	*rc_ste
	ora	#LD_exe	; set EXECUTE flag in STATE
	orb	#LD_mdd	; set MD-DIR flag in STATE
	std	*rc_ste
	ldd	*AC
	std	,u
dcaxit:	ldd	#0
	std	*AC
	jmp	NextI



   ; Get the address referenced by the instruction.
   ;
   ;   call   : MD holds current instruction
   ;		 <rc_ste> STATE bits [FETCH] = set and [DEFER] = clear
   ;   return : [D] = effective address
   ;            If AND, TAD, ISZ or DCA an indirect will be from the
   ;            data field. Set U for use by the above instructions.
   ;            indirect: U:= [WrkDp]  //  direct: U:=[WrkIp]
   ;		 STATE bit [FETCH] = clear and [DEFER] = set if indirect access
   ;   algorithm
   ;   If the I and Z bits indicate indirect page 0 and the location is (octal)
   ;   010 through 017, the location is a pre-increment autoindex access.
   ;   In this condition, the DATA-CONT light is also turned on.
   ;
   ;  11  10  9   8   7   6   5   4   3   2   1   0
   ;  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ 
   ; |   |   |   |   |   |   |   |   |   |   |   |   |  i - indirect bit [0=direct,
   ; |_ _|__ |__ |__ |__ |___|___|__ |__ |__ |__ |__ |                    1=indirect]
   ; :           :   :   :                           :  z - page bit [0=page zero, 
   ; :  opcode   : i : z :           address         :                1=current page]
   ;                                                    address - 12-bit word in page

GetAdr:	ldd	*MD
	tstb			; test for page 0
	bmi	CurPg
	lsra			; test for indirect
	bcs	Pg0Ind
	lda	*rc_ste		; page 0, direct
	anda	#LD_fet^0xFF	; => clear FETCH bit in STATE (DEFER is clear)
	sta	*rc_ste
	clra			; set PC for page 0
	ldu	*WrkIp
	rts
Pg0Ind:	lda	*rc_ste		; page 0, indirect
	ora	#LD_def		; => set DEFER bit in STATE
	anda	#LD_fet^0xFF	;    clr FETCH bit
	sta	*rc_ste
	clra
	cmpb	#0o10		; check for autoindex range
	blo	Get0I
	cmpb	#0o17
	bhi	Get0I
	sta	*Temp1
	lda	*rc_ste+1
	ora	#LD_dac		; set DATA-CONT flag in STATE
	sta	*rc_ste+1
	lda	*Temp1
	lslb			; translate to 16-bit address
	leau	d,x		; get effective address
	ldd	,u
	addd	#1		; pre-increment
	anda	#Wmask		; prevent 12-bit overflow
	std	,u
	ldu	*WrkDp
	rts
Get0I:	lslb			; translate to 16-bit address
	ldd	d,x
	ldu	*WrkDp
	rts
CurPg:	andb	#Offset
	stb	*Temp1		; save current page offset
	lsra			; check for indirect
	bcs	CurInd
	lda	*rc_ste		; current page, direct
	anda	#LD_fet^0xFF	; => clear FETCH bit in STATE (DEFER is clear)
	sta	*rc_ste
	ldd	*PC		; get current page
	andb	#LoPage
	orb	*Temp1		; insert offset
	ldu	*WrkIp
	rts
CurInd:	lda	*rc_ste		; current page, indirect
	ora	#LD_def		; => set DEFER bit in STATE
	anda	#LD_fet^0xFF	;    clr FETCH bit
	sta	*rc_ste
	ldd	*PC
	andb	#LoPage
	orb	*Temp1
	lslb			; translate to 16-bit address
	rola
	ldd	d,x		; get indirect address
	ldu	*WrkDp
	rts



    ; Go to an address.
    ; Checks for the interrupt inhibit, and if set removes it.
    ; The I and U fields must be current with the Buffer after this instruction.
    ; An instruction field change always implies in interrupt inhibit.
    ; The Time Share instruction CUF may not set the inhibit,
    ; as it stops the interrupt.

JMPsub:	lda	*IOflgs		; check for interrupt inhibit
	bpl	_dojmp
	anda	#Jdelay^0xFF	; remove inhibit
	sta	*IOflgs
	ldx	*BufIp		; update working field pointers
	stx	*WrkIp
_dojmp:	lda	*BuffUI		; update U (also does F)
	sta	*WorkUI
	bsr	GetAdr		; clears FETCH flag in STATE
	sta	*savacc
	lda	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	sta	*rc_ste
	lda	*savacc
	jmp	_savpc



; The JMS instruction stores the return address in relative word zero of
; the subroutine, and the execution starts at relative word one.
; Subroutine return is done with an indirect JMP through the return address.
; Subroutines commonly increment their return addresses to index through
; inline parameter lists or to perform conditional skips over instructions
; following the call.

    ; Jump to a subroutine.
    ; Stores the return address at the subroutine entry point and continues at
    ; the next instruction.
    ; Interrupt inhibit and U, F and D handling as in the above JMP instruction.

JMSsub:	lda	*IOflgs
	bpl	_dojms
	anda	#Jdelay^0xFF
	sta	*IOflgs
	ldx	*BufIp
	stx	*WrkIp
_dojms:	lda	*BuffUI
	sta	*WorkUI
	jsr	GetAdr		; clears FETCH flag in STATE
	std	*Temp2
	ldd	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	orb	#LD_mdd		; set MD-DIR flag in STATE
	std	*rc_ste
	ldd	*Temp2
	lslb
	rola
	leau	d,x
	ldd	*PC
	addd	#1
	anda	#Wmask
	std	,u		; store return address
	ldd	*Temp2
	jmp	_nxtni



; The IOT instruction has the following form.
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;        |     |           |     |
;        |     |   device  | op  |
;
; The IOT instruction specifies one of up to 8 operations on one of 64
; devices.  Typically (but not universally), each bit of the op field
; evokes an operation, and these can be microcoded in right to left
; order.  Prior to the PDP-8/E, there were severe restrictions on the
; interpretation of the op field that resulted from the fact that the
; operation was delivered as a sequence of IOP pulses, each on a separate
; line of the I/O bus.  Each line was typically used to evoke a different
; device function, so essentially, the operation 000 was always a no-op
; because it evoked no functions, and the code 111 evoked all three
; functions in series.
;
; As an example of the use of IOT instructions, consider the console
; terminal interface.  On early PDP-8 systems, this was always assumed to
; be an ASR 33 teletype, complete with low-speed paper tape reader and
; punch.  It was addressed as devices 03 (the keyboard/reader) and 04
; (the teleprinter/punch):
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;              |0 0 0 0 1 1|0 0 1  - KSF - keyboard skip if flag
;              |0 0 0 0 1 1|0 1 0  - KCC - keyboard clear flag
;              |0 0 0 0 1 1|1 0 0  - KRS - keyboard read static
;
; The keyboard flag is set by the arrival of a character.  The KCC
; instruction clears both the flag and the accumulator. KRS ors the 8 bit
; input data with the low order 8 bits of AC.  The commonly used KRB
; instruction is the or of KCC and KRS.  To await one byte of input, use
; KSF to poll the flag, then read the byte with KRB.
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;              |0 0 0 1 0 0|0 0 1  - TSF - teleprinter skip if flag
;              |0 0 0 1 0 0|0 1 0  - TCF - teleprinter clear flag
;              |0 0 0 1 0 0|1 0 0  - TPC - teleprinter print static
;
; The teleprinter flag is set by the completion of the TPC operation (as
; a result, on startup, many applications output a null in order to get
; things going).  TCF clears the flag, and TPC outputs the low order 8
; bits of the accumulator.  The commonly used TLS instruction is the or
; of TCF and TPC.  To output a character, first use TSF to poll the flag,
; then write the character with TLS.
;
; IOT instructions may be used to initiate data break transfers from block
; devices such as disk or tape.  The term "data break" was, for years,
; DEC's preferred term for cycle-stealing direct-memory-access data
; transfers.
;
; Some CPU functions are accessed only by IOT instructions.  For example,
; interrupt enable and disable are IOT instructions:
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;              |0 0 0 0 0 0|0 0 1  - ION - interrupts turn on
;              |0 0 0 0 0 0|0 1 0  - IOF - interrupts turn off
;
; An interrupt is requested when any device raised its flag.  The console
; master clear switch resets all flags and disables interrupts.  In
; effect, an interrupt is a JMS instruction to location zero, with the
; side effect of disabling interrupts.  The interrupt service routine
; is expected to test the device flags and perform the operations needed
; to reset them, and then return using ION immediately before the indirect
; return JMP.  The effect of ION is delayed so that interrupts are not
; enabled until after the JMP.
;
; The instructions controlling the optional memory management unit are
; also IOT instructions.  This unit allows the program to address up to
; 32K of main memory by adding a 3 bit extension to the memory address.
; Two extensions are available, one for instruction fetch and direct
; addressing, the other for indirect addressing.


    ; Will use the middle 6 bits of the instruction as the device address
    ; and the last 3 bits to indicate the action to be taken by the device.
    ; Access the device handler with the IOT code in A.
    ; Codes above 7 are "impossible" and used for special access.
    ; In this case the handler will exit with a status code in A.

IOTsub:	tst	*WorkUI	; in User Mode?
	lbmi	U_Mode	; y - User interrupt, do No-OPeration
	ldu	#Device
	ldd	*MD
	lsra
	rorb
	rorb
	andb	#IOTmask	; [B] = device (*2 for table indexing)
	lda	*MD+1
	anda	#0b111		; [A] = IOT code
	jmp	[b,u]



; OPR routine.
; One of the most complex due to the microcoded execution.
; A wide variety of operations are available through the OPR
; microcoded instructions:
;          _ _ _ _ _ _ _ _ _ _ _ _
; Group 1 |1|1|1|0|_|_|_|_|_|_|_|_|
;                  1                - CLA - clear AC
;                    1              - CLL - clear the L bit
;                      1            - CMA - ones complement AC
;                        1          - CML - complement L bit
;                                1  - IAC - increment <AC>
;                          1 0 0    - RAR - rotate <L,AC> right
;                          0 1 0    - RAL - rotate <L,AC> left
;                          1 0 1    - RTR - rotate <L,AC> right twice
;                          0 1 1    - RTL - rotate <L,AC> left twice
;                          0 0 1    - 6-bit "byte" swap AC (on pdp8/e)
;
; In general, the above operations can be combined by oring the bit
; patterns for the desired operations into a single instruction.
; If none of the bits are set, the result is the NOP instruction.
; When these operations are combined, they operate top to bottom in the
; order shownabove. The exception to this is that IAC cannot be combined
; with the rotate operations on some models, and attempts to combine
; rotate operations have different effects from one model to another
; (for example, on the PDP-8/E, the rotate code 001 means swap 6 bit
; bytes in the accumulator (BSW), while previous models took this to
; mean something like "shift neither left nor right 2 bits").
;
;          _ _ _ _ _ _ _ _ _ _ _ _
; Group 2 |1|1|1|1|_|_|_|_|_|_|_|0|
;                    1     0        - SMA - skip on AC < 0      \
;                      1   0        - SZA - skip on AC = 0       > OR group
;                        1 0        - SNL - skip on L != 0      /
;                    0 0 0 1        - SKP - skip unconditional
;                    1     1        - SPA - skip on AC >= 0     \
;                      1   1        - SNA - skip on AC != 0      > AND group
;                        1 1        - SZL - skip on L = 0       /
;                  1                - CLA - clear AC
;                            1      - OSR - or switches with AC
;                              1    - HLT - halt
;
; The above operations may be combined by oring them together, except
; that there are two distinct incompatible groups of skip instructions.
; When combined, SMA, SZA and SNL, skip if one or the other of the
; indicated conditions are true (logical or), while SPA, SNA and SZL skip
; if all of the indicated conditions are true (logical and).  When combined,
; these operate top to bottom in the order shown; thus, the accumulator may
; be tested and then cleared. Setting the halt bit in a skip instruction is
; a crude but useful way to set a breakpoint for front-panel debugging.
; If none of the bits are set, the result is an alternative form of no-op.
;
;          _ _ _ _ _ _ _ _ _ _ _ _
; Group 3 |1|1|1|1|_|_|_|_|_|_|_|1|
; 
; A third group of operate microinstructions (with a 1 in the least
; significant bit) deals with the optional Extended Arithmetic Element to
; allow such things as hardware multiply and divide, 24 bit shift
; operations, and normalize.  These operations involve an additional data
; register, MQ or multiplier quotient, and a small step count register.
; On the PDP-8/E and successors, MQ and the instructions for loading and
; storing it were always present, even when the EAE was absent, and the
; EAE was extended to provide a useful variety of 24 bit arithmetic
; operations.

OPRsub:	lda	*rc_ste
	ora	#LD_exe		; set EXECUTE bit in STATE
	sta	*rc_ste
	ldd	*MD
	rora			; test for group 1
	bcs	chkpg2

  ; OPR group 1

	bitb	#0b10000000
	beq	_cll
	clra			; CLA
	sta	*AC
	sta	*AC+1
_cll:	bitb	#0b01000000
	beq	_cma
	clr	*Link		; CLL
_cma:	bitb	#0b00100000
	beq	_cml
	ldd	*AC		; CMA
	coma
	anda	#ACmask
	comb
	std	*AC
_cml:	ldb	*MD+1
	bitb	#0b00010000
	beq	_iac
	lda	*Link		; CML
	coma
	anda	#Lmask
	sta	*Link
_iac:	bitb	#0b00000001
	beq	rotate
	ldd	*AC		; IAC
	addd	#1
	anda	#ACmask
	std	*AC
rotate:	ldb	*MD+1
	andb	#0b00001110 	; isolate rotate and swap bits
	ldu	#Group1
	jmp	[b,u]

  ; OPR group 2

chkpg2:	bitb	#0b00000001
	bne	chkpg3
	bitb	#0b01000000
	beq	_skp1
	lda	*AC		; SMA
	bita	#0b00001000
	bne	SkpTst
_skp1:	bitb	#0b00100000
	beq	_skp2
	ldd	*AC		; SZA
	beq	SkpTst
_skp2:	ldb	*MD+1
	bitb	#0b00010000
	beq	_skp3
	lda	*Link		; SNL
	bne	SkpTst
_skp3:	bitb	#0b00001000
	beq	_skp4
Skip:	ldd	*PC
	addd	#1
	anda	#ACmask
	std	*PC
	bra	_skp4
SkpTst:	ldb	*MD+1
	bitb	#0b00001000
	beq	Skip
_skp4:	ldb	*MD+1
	bpl	_skp5
	ldd	#0		; CLA
	std	*AC


    ; test for OSR and HLT. Start with a fast check.

_skp5:	ldb	*MD+1		; fast exit check
	bitb	#0b00000110 	; OSR|HLT?
	lbeq	NextI
	tst	*WorkUI		; check for OSR/HLT block
	lbmi	U_Mode
_skp6:	bitb	#0b00000100
	beq	_skp7
	ldd	*AC		; OSR
	ora	*SR
	orb	*SR+1
	std	*AC
	ldb	*MD+1
_skp7:	bitb	#0b00000010
	lbeq	NextI
	.if	RealCons
	clr	*conact		; HALT --> clear [ACTrun]
	.else
	ldb	*MONflg		; HLT
	orb	#Halt		; -> causes RUN lamp to go OFF
	stb	*MONflg
	.endif
	jsr	Errpnt		; ** should jump to the monitor
	.fdb	Hltmsg		; ** this HALTs the simulator too !!


  ; OPR group 3

chkpg3:	bitb	#0b10000000
	beq	chkmqa
	clra
	sta	*AC		; CLA
	sta	*AC+1
chkmqa:	andb	#0b01010000 	; isolate MQA and MQL bits
	lbeq	NextI
	cmpb	#0b01000000
	bne	chkmql
	ldd	*AC		; MQA
	ora	*MQ
	orb	*MQ+1
	std	*AC
	jmp	NextI
chkmql:	cmpb	#0b00010000
	bne	_swp
	ldd	*AC		; MQL
	std	*MQ
	ldd	#0
	std	*AC
	jmp	NextI
_swp:	ldy	*AC		; SWP
	ldd	*MQ
	std	*AC
	sty	*MQ
	jmp	NextI



BSW:	ldd	*AC
	rolb		; upper and lower bytes in A and B
	rola
	rolb
	rola
	exg	a,b	; swap bytes
	lsra		; align A
	lsra
	rolb		; align B
	rolb
	lsra		; align AC
	rorb
	lsra
	rorb
	std	*AC
	jmp	NextI

ROL:	lda	*Link
	adda	#LtoC	; move Link to carry
	ldd	*AC	; does not change carry
_left:	rolb		; rotate AC and Link (carry) left
	rola
	sta	*Link
	anda	#ACmask	; save AC
	std	*AC
	lda	*Link	  ; save Link
	anda	#Lmask
	sta	*Link
	jmp	NextI

RTL:	lda	*Link
	adda	#LtoC
	ldd	*AC
	rolb
	rola
	adda	#LtoC	; just affects the Link bit
	bra	_left

ROR:	ldd	*AC
	ora	*Link
_right:	lsra
	rorb
	std	*AC
	rora		; carry to Link "the easy way"
	lsra
	lsra
	lsra
	sta	*Link
	jmp	NextI

RTR:	ldd	*AC
	ora	*Link
	lsra
	rorb
	bcc	_right
	ora	#Lmask
	bra	_right

	.if	0
RORROL: ldd	*AC	; start with ROR
	ora	*Link
	lsra
	rorb
	std	*Temp2	; first shift to MD
	lda	*Link	; now ROL
	adda	#LtoC	; move Link to carry
	ldd	*AC	; does not change carry
	rolb		; rotate AC and Link (carry) left
	rola
	sta	*Link
	anda	#ACmask	; MD and AC to AC
	anda	*Temp2
	andb	*Temp2+1
	std	*AC
	lda	*Link	; save Link
	anda	#Lmask
	sta	*Link
	jmp	NextI
	.endif

ROLROR:	lda	*Link	; start with ROL
	adda	#LtoC	; move Link to carry
	ldd	*AC	; does not change carry
	rolb		; rotate AC and Link (carry) left
	rola
	anda	#ACmask	; save AC
	std	*Temp2
	ldd	*AC	; Now do ROR
	ora	*Link
	lsra
	rorb
	anda	*Temp2	; MD and AC to AC
	andb	*Temp2+1
	std	*AC
	rora		; carry to Link "the easy way"
	lsra
	lsra
	lsra
	sta	*Link
	jmp	NextI


    ; default "no device" IOT
    ; if the device does not exist, there is no action

Nodev	=	NextI

    ; NOTE: the following was the previous inplementation for "no device"

;Nodev:	jsr	Errpnt
;	.fdb	NoIOT
;NoIOT:	.asciz	"Access to non-existent I/O device"




    ; Rotate and swap routines for group 1 OPR.
    ; NOTE: The "undefined" instructions may have rotated the AC and
    ;       Link into the MD register and then in the opposite direction into itself.
    ;       Next the AC and MD registers may have been anded and put into the AC.
    ;       State of the link would depend on which direction was done first.

Group1:	.fdb	NextI	; No operation
	.fdb	BSW	; Byte swap AC
	.fdb	ROL	; Rotate AC and Link left
	.fdb	RTL
	.fdb	ROR	; Rotate AC and Link right
	.fdb	RTR
	.fdb	ROLROR	; ROR-ROL (undefined on E)
	.fdb	Illins	; RTR-RTL (undefined on E


    ; Device routine address array. 64 Possible devices.
    ; NOTE: The KM8-E is done but not tested.

Device:	.fdb	IOT00	; Device 00 Interrupts (basic system)
	.fdb	Nodev	; Device 01.
	.fdb	Nodev	; Device 02.
	.fdb	IOT03	; Device 03 KL8-E Teletype keyboard/reader
	.fdb	IOT04	; Device 04.  "      "     teleprinter/punch
	.fdb	Nodev	; Device 05.
	.fdb	Nodev	; Device 06.
	.fdb	Nodev	; Device 07.
	.fdb	Nodev	; Device 10 KP8-E Power Fail Detect
	.fdb	Nodev	; Device 11.
	.fdb	Nodev	; Device 12.
	.fdb	Nodev	; Device 13 DK8-EA/EC/EP Real Time Clocks
	.fdb	Nodev	; Device 14.
	.fdb	Nodev	; Device 15.
	.fdb	Nodev	; Device 16.
	.fdb	Nodev	; Device 17.
	.fdb	IOT20	; Device 20 KM8-E Memory ext. & Time Share
	.fdb	IOT21	; Device 21   "      "    "   "   "    "
	.fdb	IOT22	; Device 22   "      "    "   "   "    "
	.fdb	IOT23	; Device 23   "      "    "   "   "    "
	.fdb	IOT24	; Device 24   "      "    "   "   "    "
	.fdb	IOT25	; Device 25   "      "    "   "   "    "
	.fdb	IOT26	; Device 26   "      "    "   "   "    "
	.fdb	IOT27	; Device 27   "      "    "   "   "    "
	.fdb	Nodev	; Device 30.
	.fdb	Nodev	; Device 31.
	.fdb	Nodev	; Device 32.
	.fdb	Nodev	; Device 33.
	.fdb	Nodev	; Device 34.
	.fdb	Nodev	; Device 35.
	.fdb	Nodev	; Device 36.
	.fdb	Nodev	; Device 37.
	.fdb	Nodev	; Device 40.
	.fdb	Nodev	; Device 41.
	.fdb	Nodev	; Device 42.
	.fdb	Nodev	; Device 43.
	.fdb	Nodev	; Device 44.
	.fdb	Nodev	; Device 45.
	.fdb	Nodev	; Device 46.
	.fdb	Nodev	; Device 47.
	.fdb	Nodev	; Device 50.
	.fdb	Nodev	; Device 51.
	.fdb	Nodev	; Device 52.
	.fdb	Nodev	; Device 53.
	.fdb	Nodev	; Device 54.
	.fdb	Nodev	; Device 55.
	.fdb	Nodev	; Device 56.
	.fdb	Nodev	; Device 57.
	.fdb	Nodev	; Device 60.
	.fdb	Nodev	; Device 61.
	.fdb	Nodev	; Device 62.
	.fdb	Nodev	; Device 63.
	.fdb	Nodev	; Device 64.
	.fdb	Nodev	; Device 65.
	.fdb	Nodev	; Device 66.
	.fdb	Nodev	; Device 67.
	.fdb	Nodev	; Device 70.
	.fdb	Nodev	; Device 71.
	.fdb	Nodev	; Device 72.
	.fdb	Nodev	; Device 73.
	.fdb	Nodev	; Device 74.
	.fdb	Nodev	; Device 75.
	.fdb	Nodev	; Device 76.
	.fdb	Nodev	; Device 77.


    ; Interrupt IOT module.  Comes with the basic system.
    ;
    ; Standard IOPs' are 0 through 7.  Higher ones are "impossible",
    ; and used by the emulator for system initiation (8) and reset (9).

IOT00:	ldu	#IOP00	; use an opcode array
	lsla		; get the IOP code
	jmp	[a,u]


IOP00:	.fdb	SKON	; 0
	.fdb	ION	; 1
	.fdb	IOF	; 2
	.fdb	SRQ	; 3
	.fdb	GTF	; 4
	.fdb	RTF	; 5
	.fdb	IOT0_6	; 6
	.fdb	CAF	; 7
	.fdb	No_op	; 8 = Initialise
	.fdb	No_op	; 9 = Reset


    ; Skip if interrupt on and turn off.
    ; Assume "on" not influenced by ION and JMP/JMS inhibits.

SKON:	ldd	#Doinst
	cmpd	*emufnc	; is the processing routine?
	lbeq	NextI	; y -
	std	*emufnc	; n - make it so! (set vector to processing routine)
	jmp	PCskip


    ; Turn interrupt on. Actual response delayed till after the next instruction
    ; (unless are already on).
    ; NOTE: This is independent of JMP/JMS inhibit.

ION:	ldd	*emufnc	; get current selected processing routine
	cmpd	#Doinst
	lbne	NextI
	lda	#2
	sta	*Idelay
	ldd	#Chkdly
setjmp:	std	*emufnc	; set vector to processing routine
	jmp	NextI

    ; Turn interrupts off (unconditional).
    ; NOTE: Does not change JMP/JMS inhibit.

IOF:	ldd	#Doinst
	bra	setjmp

    ; Skip if interrupt request

SRQ:	lda	*IOflgs
	anda	#Jdelay^0xFF ; drop JMP/JMS inhibit
	ora	*IOflgs+1
	lbne	PCskip
	jmp	NextI

    ; Get flags

GTF:	ldb	*WorkUI	; get I field and user state
	bpl	getInt	; test for User mode
	orb	#0b00001000 ; set U bit
getInt:	anda	#0b00001111 ; keep U and F bits
	ldu	*emufnc	; check interrupt status
	cmpu	#Doinst
	beq	_getD
	orb	#0b00010000 ; set Ion bit
_getD:	lslb		   ; move
	lslb
	lslb
	orb	*WorkD	   ; add in D field
	lda	*IOflgs	   ; check for interrupt req
	anda	#Jdelay^0xFF  ; drop JMP/JMS delay
	ora	*IOflgs+1
	bne	_setI
	clra
	bra	getBlk
_setI:	lda	#0b00000010   ; put in IRQ
getBlk:	tst	*IOflgs	     ; check for interrupt block
	bpl	getLnk
	ora	#0b00000001   ; put in JMP/JMS delay
getLnk:	tst	*Link
	beq	savFlg
	ora	#0b10000000   ; put in Link
savFlg:	std	*AC
	jmp	NextI

    ; Restore flags

RTF:	ldd	*AC	; restore Link
	rola
	anda	#Lmask
	sta	*Link
	andb	#Dbits	; restore D
	stb	*WorkD
	lslb
	ldu	#GetFld
	ldd	b,u
	std	*WrkDp
	lda	*AC+1	; restore U and I (to Buffer)
	lsra
	lsra
	lsra
	anda	#0b00001111
	adda	#0b01111000  ; shift U to MSB
	anda	#Ubit|Fbits
	sta	*BuffUI
	lsla		    ; drops U
	ldd	a,u
	std	*BufIp
	ldd	#Chkint	    ; Interrupts on.
			    ; The JMP/JMS delay takes care of the 1 cycle delay
	std	*emufnc	    ; set vector to processing routine
	bra	_setJ	    ; set JMP/JMS delay

    ; Clear all Flags

CAF:	jsr	Clear
	jmp	NextI


IOT0_6:	jsr	Errpnt
	.fdb	NoIerr


    ; Use for bypassing IOT device setup or initialisation.

No_op:	clra		; no error
	rts		; do not do setup or initialise



    ; KM8-E Memory Extension and Time-Share Option.
    ;
    ; Memory extension allows up to 8 fields (4096 words/field).
    ; Time share has Executive and User modes, controlled by the U bit.
    ; If it is 1 the computer is in time share mode and all OSR, HLT and
    ; IOT instructions are ignored and a time share interrupt is generated.
    ; The interrupt sets the Instruction and Data fields and the time share bit to 0.

IOT20:	ldb	#0*2	; field number times 2
	bra	km8iot
IOT21:	ldb	#1*2
	bra	km8iot
IOT22:	ldb	#2*2
	bra	km8iot
IOT23:	ldb	#3*2
	bra	km8iot
IOT24:	ldb	#4*2
	bra	km8iot
IOT25:	ldb	#5*2
	bra	km8iot
IOT26:	ldb	#6*2
	bra	km8iot
IOT27:	ldb	#7*2
km8iot:	lsla		; IOT times 2
	ldu	#mm8op0
	jmp	[a,u]

mm8op0:	.fdb	IOT2N	; 0
	.fdb	CDF	; 1
	.fdb	CIF	; 2
	.fdb	CDFCIF	; 3
	.fdb	mm8_N4	; 4
	.fdb	IOT2N	; 5
	.fdb	IOT2N	; 6
	.fdb	IOT2N	; 7
	.fdb	No_op	; 8
	.fdb	No_op	; 9


mm8_N4:	ldu	#mm8op1
	jmp	[b,u]

mm8op1:	.fdb	CINT
	.fdb	RDF
	.fdb	RIF
	.fdb	RIB
	.fdb	RMF
	.fdb	SINT
	.fdb	CUF
	.fdb	SUF

CDF:	ldu	#GetFld
	ldu	b,u
	stu	*WrkDp
	lsrb
	stb	*WorkD
	jmp	NextI

CIF:	ldu	#GetFld	; must have I field
	ldu	b,u
	lbeq	Flderr
	lsrb
_cif0:	stu	*BufIp
	tst	*BuffUI	; keep Time Share Mode
	beq	_cif1
	orb	#Ubit
_cif1:	stb	*BuffUI
_setJ:	lda	#Jdelay	    ; set JMP/JMS delay
	ora	*IOflgs
	sta	*IOflgs
	jmp	NextI

CDFCIF:	ldu	#GetFld
	ldu	b,u
	lbeq	Flderr	; must have I field
	stu	*WrkDp
	lsrb
	stb	*WorkD
	bra	_cif0

RDF:	ldb	*WorkD
_rdf0:	lsla
	lsla
	lsla
_rdf1:	orb	*AC+1
	stb	*AC+1
	jmp	NextI

RIF:	ldb	*WorkUI
	bra	_rdf0

RIB:	ldb	*SaveUI
	lslb
	lslb
	lslb
	orb	*SaveD
	tst	*SaveUI
	beq	_rdf1
	orb	#0b01000000 ; set U mode
	bra	_rdf1

RMF:	ldd	*SaveUI
	sta	*BuffUI
	stb	*WorkD
	ldd	*SavDp
	std	*WrkDp
	ldd	*SavIp
	std	*BufIp
	bra	_setJ

CINT:	lda	*IOflgs
	anda	#UserI^0xFF
	sta	*IOflgs
	jmp	NextI

SINT:	lda	*IOflgs
	anda	#UserI
	lbne	PCskip
	jmp	NextI

    ; Set Time Share mode to Executive.
    ; NOTE: Not clear if this sets the JMP/JMS delay,
    ;       as it stops Time Share interrupts.

CUF:	lda	*BuffUI
	anda	#Ubit^0xFF
	sta	*BuffUI
	jmp	NextI

SUF:	lda	*BuffUI
	ora	#Ubit
_suf0:	sta	*BuffUI
	bra	_setJ

    ; Time share interrupt.
    ; User mode and LAS, HLT or IOT instruction.
    ; Instruction treated as a no-op and interrupt generated.

U_Mode:	lda	*IOflgs	; user interrupt in first byte
	ora	#UserI
	sta	*IOflgs
	jmp	NextI

    ; KM8-E decode error

IOT2N:	jsr	Errpnt
	.fdb	KM8Eerr


    ; KL8-E Console Teleprinter Control Module.
    ;
    ; Includes both IOT 03 and IOT 04. Interrupt request enable is common to both,
    ; and set on at reset. Also controlled by AC11 on a KIE instruction.
    ; If the bit is 0, interrupt requests are disabled.

IOT03:	ldu	#IOP03	; use an opcode array
	bra	iotdec


IOT04:	ldu	#IOP04
iotdec:	lsla		; get the IOP code
	jmp	[a,u]

IOP03:	.fdb	KCF	; 0
	.fdb	KSF	; 1
	.fdb	KCC	; 2
	.fdb	IOT033	; 3
	.fdb	KRS	; 4
	.fdb	KIE	; 5
	.fdb	KRB	; 6
	.fdb	IOT037	; 7
	.fdb	No_op	; 8 Initialise
	.fdb	No_op	; 9 Reset

IOP04:	.fdb	TFL	; 0
	.fdb	TSF	; 1
	.fdb	TCF	; 2
	.fdb	IOT043	; 3
	.fdb	TPC	; 4
	.fdb	TSK	; 5
	.fdb	TLS	; 6
	.fdb	IOT047	; 7
	.fdb	TTinit	; 8 Initialise
	.fdb	TTrst	; 9 Reset


    ; RS_ion: enable RS232 input interrupt

RS_ion:	lda	ACIADA	; clear possible pending ACIA interrupt
	lda	#ACIAI	; enable ACIA-IRQ, clock/16, 8N2
	sta	ACIACS	; RTS=0 (LED off)
	.if	RealCons
	orcc	#0x40	; disable FIRQ: mutual excl. access to rtssts
	sta	*rtssts
	andcc	#0xAF	; enable FIRQ _and_ IRQ (RS232)
	.else
	andcc	#0xEF	; enable IRQ
	.endif
	rts



    ; Initialise the TTY handler

TTinit:	jsr	GetFlg	; get the interrupt bit
	std	*Ibit
	coma		; Interrupt bit complement
	comb
	std	*Icom
	ldd	#TTIchr	; set up the input interrupt vector
	std	.irq
	bsr	RS_ion	; turn the interrupt on
TTrst:	ldd	#0	; clear the device flags
	std	*TTIflg
	inca		; set up the interrupt status
	sta	*Istat
	ldd	*IOflgs
	anda	*Icom
	andb	*Icom+1
	std	*IOflgs
	andcc	#0xEF	; turn IRQ on
	orcc	#4	; ccZ=1 :: successful setup
	rts



    ; Character input handler, driven by the IRQ interrupt.
    ; Check for ^S to call the monitor (if RealConsole is not connected).

TTIchr:	lda	ACIADA	; get serial input data / clear IRQ
	.if	RealCons
	bsr	ttisub	; put character in system
	rti
	.else
	cmpa	#XOFF	; turn on single step (^S)?
	beq	ttstep
	bsr	ttisub	; put character in system
	rti
ttstep:	lda	*MONflg	; call monitor
	ora	#Console
	sta	*MONflg
	rti
	.endif


ttisub:	ora	#0x80	; add in for TTY compatibility
	sta	*TTIdat
	lda	#1	; set keyboard\reader flag
	sta	*TTIflg
	tst	*Istat	; update interrupt request
	beq	charxt
	ldd	*IOflgs	; Interrupt was enabled
	ora	*Ibit
	orb	*Ibit+1
	std	*IOflgs	; request interrupt
charxt:	rts


    ; Update the interrupt request.
    ; If interrupt enabled (and a flag set) request an interrupt.
    ; If interrupt disabled clear the request.

Isetup:	tst	*Istat	; Interrupt enabled?
	beq	Iclr
	ldd	*TTIflg	; any requests?
	beq	Iclr
Iset:	ldd	*IOflgs	; request interrupt
	ora	*Ibit
	orb	*Ibit+1
	bra	isetxt
Iclr:	ldd	*IOflgs	; clear interrupt request
	anda	*Icom
	andb	*Icom+1
isetxt:	std	*IOflgs
	jmp	NextI


    ; TTI will use the IRQ interrupt to get a character from the keyboard.
    ; The IRQ interrupt will be enabled to trap character input to a handler
    ; that will set the Initial baud time to correct for servicing delay.
    ; On exit the timing will be reset.
    ; TTO will not turn the interrupt off while printing characters. This
    ; could chop an output character up, making it look like a link exit.
    ; In the future the NMI could be used to output a character, followed
    ; by a returned character. The characters could either be NULL or data.

    ; ***  TTI section (Teletype keyboard or paper tape reader)  ***

    ; Clear keyboard flag.
    ; Does not enable the paper tape reader to get another character.

KCF:	bra	kccxit

KSF:	tst	*TTIflg	; Skip on keyboard flag
ksfskp:	lbne	PCskip
	jmp	NextI


    ; Clear the keyboard flag.
    ; Enables the paper tape reader to get another character.
    ; NOTE: DEC modifies the standard ASR-33 Teletype to input
    ;       a single character from the reader on each request.
    ;       It is assumed that a keyboard character always sets the flag.

KCC:	ldd	#0
kccset:	std	*AC
; --	jsr	RDRrun	; NOTE: routine does not yet exist
kccxit:	clr	*TTIflg
	bra	Isetup


KRS:	ldd	*AC	; read keyboard buffer static
	orb	*TTIdat
	std	*AC
	jmp	NextI

KIE:	ldd	*AC	; set/clear interrupt enable
	andb	#1
	stb	*Istat
	bra	Isetup

KRB:	clra		; read keyboard buffer dynamic
	ldb	*TTIdat
	bra	kccset

			; TTI error message(s)
IOT033:
IOT037:	jsr	Errpnt
	.fdb	TTI_3_7



    ; ***  TTO section (Teletype printer or paper tape punch)  ***

TFL:	lda	#1	; set the teleprinter flag
	sta	*TTOflg
	bra	Isetup

TCF:	clr	*TTOflg	; Clear the teleprinter flag
	bra	Isetup

			; Load teleprinter and print.
			; Does not clear the Teleprinter flag.
TPC:	lda	*AC+1
	anda	#0x7F	; drop high bit
	jsr	outch
	lda	#1
	sta	*TTOflg
	bra	Isetup

TSF:	tst	*TTOflg	; Skip on teleprinter flag
	bra	ksfskp

TSK:	lda	*TTIflg	; Skip on printer of keyboard flag
	ora	*TTOflg	;   if the interrupt is enabled
	anda	*Istat
	lbeq	NextI
	jmp	PCskip


    ; Load teleprinter sequence.
    ; Same as TPC, except that the Teleprinter flag is cleared first.
    ; Since the character output is not interrupt driven the effect is
    ; the same as TPC.

TLS:	bra	TPC	; no difference

			; TTO error message(s)
IOT043:
IOT047:	jsr	Errpnt
	.fdb	TTO_3_4




; Binary program loader. Field and origin will be set up to field 0
; and PC=0200.  Data must start and end with leader / trailer code.
; An optional ASCII text before the leader is accepted, and printed.
; The trailer can be followed by nulls.  The first character after the
; trailer/nulls is passed back to the caller. Rub-out character is not
; processed.  An error is flagged by <mumble>.
; Program PC and Field will have the exit values from the loader,
; which serves as a default starting address.
;
;   Description of binary loader format
; ---------------------------------------
; Two characters representing the address (origin) into which the first
; command on the next portion of the tape will be placed. Successive commands
; are placed in memory at addresses: origin+1, origin+2, etc.
; The initial character of the origin has no punch in channel 8, while
; channel 7 is punched. The second character designating the origin has
; no punches in either channel 8 or 7.
; A concluding 2-character group represents the checksum with no punches present
; in channel 8 or channel 7.
; When channel 8 and channel 7 are punched, the character is a "field" setting:
; 1 1 x x x 0 0 0  indicate the memory field into which the following data is
; to be loaded. If, for example "x x x" is "1 0 1", all data following the field
; designator should be loaded into memory field five.
; Contents have no punch in channel 7 or 8.
; The papertape leader and trailer contains any code with channel 8 punched.
; Octal 200 is preferred (several times repeated).
;
;    8 7  6 5 4 s 3 2 1
;   --------------------
;    0 0  x x x . x x x    --> data byte
;    0 1  x x x . x x x    --> origin code, plus next following byte
;    1 0  x x x . x x x    --> leader/trailer byte, [6-1] prefered 0
;    1 1  x x x . 0 0 0    --> field code (single byte)
;
;
;       tape channel        channel 8 and 7    (s = sprocket hole)
;    8 7  6 5 4 s 3 2 1      indicate :
;   ---------------------------------------------------------------------
;    1 0  0 0 0 . 0 0 0     leader code
;
;    0 1  0 0 0 . 0 1 0     origin             octal origin 0200
;    0 0  0 0 0 . 0 0 0
;    0 0  1 1 1 . 0 1 0     contents of 0200   command 7200 or CLA
;    0 0  0 0 0 . 0 0 0
;    0 0  0 1 1 . 0 1 0     contents of 0201   command 3276 or DCA Z 076
;    0 0  1 1 1 . 1 1 0
;    0 0  1 1 1 . 1 0 0     contents of 0202   command 7402 or HLT
;    0 0  0 0 0 . 0 1 0
;    0 0  0 0 0 . 1 0 0     checksum           checksum = 0422
;    0 0  0 1 0 . 0 1 0
;    1 0  0 0 0 . 0 0 0     trailer code
;
; Checksum = 102 + 000 + 072 + 000 + 032 + 076 + 074 + 002.
;
;   Entry to loader.
;   Set up starting defaults and bypass leader code

PDPbin:	jsr	RDRini	; set up interrupt-driven data reception
	ldx	GetFld	; get start of RAM, default field is 0
	clra
	clrb
	tfr	d,y	; clear byte read-in counter
	std	*binsum	; clear checksum
	std	*binsv0	; and temp save locations
	std	*binsv1
	ldd	#0o200	; default origin
	std	*PC
    ;	lda	#CtrlC
    ;	jsr	outch	; transmit ^C - start (needed?)

;   some paper tapes start with ASCII information, such as
;   - producer and product info
;   - load/start address

_asc0:	jsr	RDRdat	; get byte
	tsta		; msb set ?

	.if	BINldr_txt

	bmi	_ldr0	; y - print "blocks" line first
	cmpa	#Space	; n - check for ASCII printable
	bhs	_asc1	;     ok: print it
	lda	#'.	;     non-printable: output "."
_asc1:	jsr	outch
	bra	_asc0

_ldr0:	pshs	x
	ldx	#RDRms2
	jsr	pdatao	; ASCII header done, print "block" line
	puls	x
	bra	_ldr2
_ldr1:	jsr	Shw_LT	; show leader "." (max 3)
	jsr	RDRdat

	.else

	bpl	_asc0	; "eat" all ASCCI text
	bra	_ldr2	; -> process leader
_ldr1:	jsr	RDRdat	; get 1 byte

	.endif

;   leader section

_ldr2:	tfr	a,b
	andb	#0xC0	; check upper 2 bits, ignore low bits
	cmpb	#0o200	; leader code 02xx / 0200 (pref'd) ?
	beq	_ldr1	; y - get next byte

;   binary data loader - process first (received) byte

_bin0:	sta	*_bin3	     ; will need a copy
	ldd	*binsv1	     ; move saved data
	std	*binsv0
	ldb	*_bin3
	std	*binsv1	     ; save present data
	addd	*binsum	     ; update checksum
	std	*binsum
	lda	*_bin3
	bita	#0o300	     ; data, origin, or field ?
	beq	_bin1	     ; -> data: get 2nd byte of data

    ; field or origin setting

	anda	#0b11000000  ; -> check: origin or field ?
	cmpa	#0o300
	beq	_fset
	.if	BINldr_txt
	jsr	Shw_O	     ; show Origin data received
	.endif
	jsr	binwrd	     ; origin - read 2nd byte of origin
	std	*PC
	bra	binchk	     ; ok, continue loading

_fset:	lda	*_bin3	     ; field format
	tfr	a,b
	andb	#7	     ; bits 2-1-0 must be zero (spec)
	lbne	lderr1	     ; -> wrong, not zero!
	lsra		     ; ok, move field bits [6-5-4] -> [3-2-1]
	lsra
	lsra
	anda	#7	     ; maximum field number
	.if	BINldr_txt
	jsr	Shw_F	     ; show Field data received
	.endif
	ldu	#GetFld	     ; make field pointer
	lsla
	ldx	a,u
	lbeq	lderrf	     ; -> field does not exist!
	bra	binchk	     ; ok, continue loading

    ; get 2nd byte of data

_bin1:	bsr	binwrd	     ; read 2nd byte of data
	.if	BINldr_txt
	jsr	Shw_D	     ; show data "d" (max 3)
	.endif
	pshs	d
	ldd	*PC
	lslb		     ; form word offset
	rola
	leau	d,x
	stu	*binptr	     ; save previous data
	stu	*rc_bus	     ; show address in BUS (RealConsole)
	ldd	,u
	std	*bindat
	puls	d	     ; load new data
	std	,u
	std	*MD	     ; show data in MD (RealConsole)
	ldd	*PC	     ; update "PC"
	addd	#1
	anda	#0b00001111  ; prevent PC overflow
	std	*PC

    ; read 1st byte

binchk:	jsr	RDRdat
	tfr	a,b
	andb	#0xC0	     ; check upper 2 bits, ignore low bits
	cmpb	#0o200	     ; trailer code 02xx / 0200 (pref'd) ?
	bne	_bin0	     ; n -> process data

;   trailer section

	ldd	*binsum	     ; y -> undo last add
	subd	*binsv0
	subd	*binsv1
	subd	[binptr]     ; update checksum
	anda	#0b00001111
	std	*binsum	     ; must be 0, else
	bne	lderr2	     ; -> checksum error!
	ldd	*bindat	     ; restore last written data
	std	[binptr]
	ldd	*PC	     ; restore PC
	subd	#1
	anda	#Wmask
	std	*PC

    ; process trailer, possibly followed by next part (multi-part tape)
    ; assumption here: trailer becomes leader of next part, next part
    ; starts with a Field or an Origin specification.
    ; Trailer end is detected by (user entry of) the Xon character.
    ; When, after the trailer data is detected, a not-trailer, not-
    ; origin or not-field data byte is read, that byte and all following
    ; bytes are skipped.
    ; Also the byte-read-in counter [Y] is not incremented. This enables
    ; checking after reading where in the file the rubbish data started.

_tlr0:
	.if	BINldr_txt
	jsr	Shw_LT	     ; show trailer "." (max 3)
	.endif
	jsr	RDRdat
	tfr	a,b
	andb	#0xC0	     ; check upper 2 bits, ignore low bits
	cmpb	#0o200	     ; trailer code 02xx / 0200 (pref'd) ?
	beq	_tlr0	     ; y -
	bitb	#0xC0	     ; n - Origin or Field specification ?
	lbne	_bin0	     ;     y - next part ->
	cmpa	#XON	     ;     ^Q Xon entry ?
	beq	_tlr2	     ;     y - done reading
_tlr1:	jsr	RDRdat	     ;     n - read rubbish, no counter update
	cmpa	#XON	     ;         "eat all" - until Xon (entry)
	bne	_tlr1
_tlr2:	orcc	#0x10	     ; disable ACIA-IRQ on 6809 CPU
	rts		     ; Xon is the return code


;   * binwrd - read 2nd byte (data or origin)
;
;     call   : <_bin3> = high 6 bits of 12-bit word
;     return :   [D]   = 12-bit word

binwrd:	jsr	RDRdat	      ; get low 6 bits
	sta	*_bin4
	bita	#0b11000000   ; check for error
	bne	binerr	      ; error - upper 2 bits <> 0
	ldd	*binsv1
	std	*binsv0
	ldb	*_bin4
	std	*binsv1
	addd	*binsum
	std	*binsum
	lda	*_bin3
	anda	#0b00111111   ; drop high bits
	clrb
	lsra
	rorb
	lsra
	rorb
	orb	*_bin4
	rts


    ; errors detected in the binary loader

lderrf:	bsr	ldrerr	; field does not exist (no memory)
	jsr	Errpnt
	.fdb	Errmem
lderr1:	bsr	ldrerr	; field specification error
	jsr	Errpnt
	.fdb	Errfld
lderr2:	bsr	ldrerr	; checksum failed
	jsr	Errpnt
	.fdb	Errcks
binerr:	bsr	ldrerr	; bad data
	jsr	Errpnt
	.fdb	Lderr


    ; print byte counter [Y] of received bytes in the loader

ldrerr:	ldx	#ldrcnt	; new line, byte count text
	jsr	pdatao
	sty	*Temp2
	ldx	#Temp2	; print counter value (in hex)
	jmp	OUT4HS




;  * BIN loader - receive data routines (interrupt driven)

RDRini:	ldx	#RDRirq
	stx	.irq	  ; set IRQ vector
	ldx	#RDRblo
	stx	*RDRin	  ; set up buffer I/O pointers
	stx	*RDRout

	.if	BINldr_txt
	clr	*RDRddd	  ; clear block indicator counters
	clr	*RDRltr
	clr	*RDRcnt	  ; # printer chars on one line
	ldx	#RDRms1
	jsr	pdatao
	.endif

	jmp	RS_ion	  ; enable RS232 input interrupt , RTS


RDRirq:	pshs	a,x	  ; save environment [A],[X]
	lda	ACIADA	  ; get data byte, clear IRQ flag
	ldx	*RDRin
	sta	,x+	  ; store data in buffer
	cmpx	#RDRbhi	  ; crossed end of buffer ?
	bne	1$
	ldx	#RDRblo	  ; y - cycle buffer ptr
1$:	stx	*RDRin
	cmpx	*RDRout	  ; simple test: buffer overrun *_next_* byte?
	bne	2$
	ldx	#RDRmsx	  ; *** ugly: text printing inside IRQ code
	jsr	pdatao	  ; ***       but this should never occur!!
2$:	puls	a,x	  ; restore environment
	rti


RDRdat:	pshs	x
	ldx	*RDRout	  ; get retrieve ptr
1$:	cmpx	*RDRin
	beq	1$	  ; wait for data
	lda	,x+	  ; get it
	cmpx	#RDRbhi	  ; crossed end of buffer ?
	bne	2$
	ldx	#RDRblo	  ; y - cycle buffer ptr
2$:	stx	*RDRout
	leay	1,y	  ; bump byte counter
	sty	*MQ	  ; show byte count in MQ (RealConsole)
	puls	x,pc


	.if	BINldr_txt

Shw_O:	pshs	a
	lda	#'O
	bsr	RDRdmp	  ; show Origin block
	clr	*RDRddd	  ; clear block counters
	clr	*RDRltr
	puls	a,pc

Shw_F:	pshs	a
	lda	#'F
	bsr	RDRdmp	  ; show Field block
	clr	*RDRddd	  ; clear block counters
	clr	*RDRltr
	puls	a,pc

Shw_LT:	pshs	a
	clr	*RDRddd
	lda	*RDRltr	  ; check # of printed "."
	cmpa	#3
	beq	1$	  ; 3 '...' printed in this block: skip
	inca
	sta	*RDRltr
	lda	#'.
	bsr	RDRdmp
1$:	puls	a,pc

Shw_D:	pshs	a
	clr	*RDRltr
	lda	*RDRddd	  ; check # of printed "d"
	cmpa	#3
	beq	1$	  ; 3 'ddd' printed in this block: skip
	inca
	sta	*RDRddd
	lda	#'d
	bsr	RDRdmp
1$:	puls	a,pc


;  * RDRdmp - dump data indicator
;
;  call   : [A] = print code
;		'.' - for leader or trailer word
;		'd' - for data word
;		'F' - for Field word
;		'O' - for Origin word
;  return : [A] destroyed

RDRdmp:	jsr	outch	  ; print char
	lda	*RDRcnt
	inca
	cmpa	#60	  ; end of line ?
	bne	1$	  ; n -
	pshs	x	  ; y - new line
	ldx	RDRms3
	jsr	pdatao
	puls	x
	clra		  ;     clear counter
1$:	sta	*RDRcnt
	rts


RDRms1:	.fcb	CR,LF
	.asciz	"- header : "
RDRms2:	.fcb	CR,LF
	.asciz	"- blocks : "
RDRms3:	.fcb	CR,LF
	.asciz	"           "

	.endif

RDRmsx:	.fcb	CR,LF
	.fcc	"! input buffer overrun"
	.fcb	CR,LF,0




    ; Pntreg - print the PDP-8 registers

Pntreg:	lda	*Link	; link bit in wrong place
	lsra		; move it
	lsra
	lsra
	lsra
	sta	*_stp2
	lda	*WorkUI	; U in MSB
	lsla		; U to carry and 0 to LSB
	rola		; U to LSB and next 4 bits= 0
	sta	*_stp3
	ldu	#_stp4	  ; flag set routine
	ldd	*emufnc	  ; check for ION
	cmpd	#Chkint
	bsr	ClrFLG
	lda	*IOflgs	  ; check for NoI
	anda	#Jdelay
	bsr	SetFLG
	lda	*IOflgs	  ; check for Ibuss
	anda	#Jdelay^0xFF
	ora	*IOflgs+1
	bsr	SetFLG
	ldx	#Stpmsg
_stp0:	bsr	pdatao
	ldu	,x++
	lda	,x+
	bsr	PrtOct
	lda	,x	; end of formatted string?
	bne	_stp1
	rts
_stp1:	cmpa	#CR	; CR/LF flag ?
	bne	_stp0
	jsr	docrlf	; y - print new line
	leax	1,x
	bra	_stp0


    ; Set location pointed to by U to 1 if entered with ccZ == 1
    ; location set to 0 otherwise. Exit with U incremented.

ClrFLG:	beq	setf
	bra	clrf

    ; Set location pointed to by U to 1 if entered with ccZ != 0
    ; location set to 0 otherwise. Exit with U incremented.

SetFLG:	beq	clrf
setf:	lda	#1
setfx:	sta	,u+
	rts
clrf:	clra
	bra	setfx



;   Error message print routine.
;   When an error is detected, the code does a JSR to this routine, and
;   the 2 bytes following the JSR is the address of the error message.

Errpnt:
; ---	orcc	#0x50		; disable 6809 FIRQ and IRQ interrupts
	jsr	docrlf
	ldx	[0,s]		; get message pointer
	bsr	pdatao
	jsr	docrlf
	bsr	Pntreg
	.if	RealCons
	leas	2,s		; clean up stack
	clr	*conact		; try to maintain RealConsole control
	jsr	docrlf
	jmp	Doinst
	.else
	bra	.		; PROGRAM HALTED :: loop forever until RESET
	.endif



;  PrtOct - print an octal number
;
;    call   : [A] = number of digits to (1 to 4)
;	      [U] = pointer to data location (2 bytes)
;    return : requested digits printed
;	      [A] and [B] destroyed

PrtOct:	sta	*_octp2	; save digit count
	sta	*_octp3	; also need character count
	ldd	,u
	std	*_octp4
_octp0:	lda	*_octp5	; get data
	anda	#7	; make into a character
	adda	#'0
	pshs	a	; save on stack
	ldd	*_octp4	; move next digit into place
	lsra
	rorb
	lsra
	rorb
	lsra
	rorb
	std	*_octp4
	dec	*_octp2
	bne	_octp0
_octp1:	puls	a
	jsr	outch
	dec	*_octp3
	bne	_octp1
	rts



    ; Print text string
    ; [X] = pointer to start of string, ended *NULL*
    ; [A] - preserved

pdatao:	pshs	a
	lda	,x+	; get char
	beq	pdatax	; done -
pdata:	jsr	outch
	lda	,x+
	bne	pdata
pdatax:	puls	a,pc



;   Monitor program load.
;   Wait for a binary program to be loaded.
;   This means the terminal data link has been established.
;   A startup message is printed and the routine exits.

PDPld:	ldx	#ldrmss	; print loader start message
	bsr	pdatao
	jsr	PDPbin	; load the binary source
	ldx	#ldrmse	; print loader finished message
	bsr	pdatao	; and ask for another file to load
;
PDPwht:	jsr	inch	; await answer
	anda	#0x5F	; make uppercase
	cmpa	#'N
	beq	PDPldx	; N -> done loading
	cmpa	#'Y
	bne	PDPwht	; not Y -> say what? -- try again
	jsr	outch	; Y -> print answer
	bra	PDPld	;      go for next file
PDPldx:	jsr	outch	; print "N" answer
	jmp	docrlf	;       and newline / RTS





;  ** formatted message string used by Pntreg
;
;  1st entry - ASCII text string, NULL ended
;  2nd entry - 2-byte start address of variable that holds the data value
;  3rd entry - 1 byte indicating the number of octal digits to be printed
;
;  after a group as described above follows either
;  - the next group with the same lay-out
;  - one byte character CR indicating a new line must be printed
;  - one byte NULL indicating end of the formatted message to print

Stpmsg:	.asciz	" Df="
	.fdb	WorkD-1
	.fcb	1
	.asciz	" AC="
	.fdb	AC
	.fcb	4
	.asciz	" L="
	.fdb	_stp2-1
	.fcb	1
	.asciz	" MQ="
	.fdb	MQ
	.fcb	4
	.asciz	" U="
	.fdb	_stp3-1
	.fcb	1
	.asciz	" ION="
	.fdb	_stp4-1
	.fcb	1
	.asciz	" NoI="
	.fdb	_stp5-1
	.fcb	1

	.asciz	" Ireq="
;	.fdb	_stp6-1
;	.fcb	1
	.fdb	IOflgs
	.fcb	5	; all but JMP/JMS delay

	.fcb	CR	; show with next state

	.asciz	" If="
	.fdb	WorkUI
	.fcb	1
	.asciz	" PC="
	.fdb	PC
	.fcb	4
	.asciz	" In="
	.fdb	MD
	.fcb	4

	.fcb	0	; end of message list






;  initRC - initialize RealCons hardware
;
;  * must be called -before- the initialization of the pdp8/e 'hardware'.
;    initRC also sets a dummy IRQ vector.
;
;    call   : -
;    return : 6821 PIA I/O initialized for I/O Board connection
;	      switch/key entry management storage initialized
;	      FIRQ vector set to process switch/key entry / output LED's
;	      FIRQ enabled on 6809, IRQ, SWI, NMI set to dummy routine
;	      input port switch/toggle mask defined
;	      output ports all cleared

	.if	RealCons
initRC: orcc	#0x40	; make sure 6809-FIRQ is disabled
	lda	#ACIANI
	sta	ACIACS
	sta	*rtssts
	lda	ACIADA	; clear ACIA
	ldx	#PDPFRQ
	stx	.firq	; load FastIRQ destination vector
	lda	#50
	sta	*timcnt	; set flash rate approx 1 Hz.
	ldx	#dumirq
	stx	.nmi	; load NMI destination vector
; ---	stx	.swi	; load SWI destination vector
	stx	.irq	; load IRQ destination vector (ACIA)

	clra		; configure PIA port A & B
	sta	PIACA	; access data direction reg A
	sta	PIADA	; default: port A == input
	ldb	#PACTLJ	; control mask set for direction: input
	stb	PIACA	; access data reg A
	stb	*piadir	; set flag: port A set to "INPUT"

	sta	PIACB	; access data direction reg B
	coma
	sta	PIADB	; always: port B == output
	ldb	#PBCTLO	; control mask set for direction: output
	stb	PIACB	; access data reg B

	ldx	#RCRAM	; start of input port shadow
	lda	#LSTIRX	; # bytes to clear
init0:	clr	,x+
	deca
	bne	init0
	clr	*conact	     ; no console action response set
	clr	*dlyexm
	clr	*dlydep
	clr	*dlyclr
	clr	*dlyal
	clr	*dlyeal
	andcc	#0b10111111  ; enable 6809-FIRQ
	rts





  ;  ***********==========    Real Console  FIRQ  routine    ==========***********

  ; The interrupt is based on a timer event that triggers the CA1 pin of the PIA.
  ; The following processes are executed inside the interrupt routine:
  ; 1. the IRQ flag of the PIA is cleared
  ; 2. <timer> is decremented. On reaching $00 the following actions are done:
  ;    - <timer> is reloaded with <timcnt>
  ;    - <rtssts> is tested/changed and the monitor LED (at the RTS pin) is toggled
  ; 3. <SWTDBCn> is decremented. On reaching $00 the following actions are done:
  ;    - <SWTDBCn> is reloaded with a fixed value.
  ;    - the physical input port [n] is read.
  ;    - the physical data is compared with the previous read data.
  ;    - SWITCH bits that are equal are transfered to the shadow input port.
  ; 4. The console switches and toggles are processed for the pdp8/e functionality.
  ; 5. Some storage locations from the simulator are read & displayed on the LEDs.

PDPFRQ:	pshs	d,x
	lda	PIADA	; dummy read: clear IRQ flag
	lda	*timer	; source == PIA [port A] - CA1 (timer)
	deca		; adjust/check monitor LED interval elapsed ?
	sta	*timer
	bne	pirq2
			; y - perform monitor LED update
pirq1:	lda	*timcnt	; reload interval time
	sta	*timer
	ldb	*rtssts	; toggle monitor LED on/off   ** preserve IRQ-enable bit **
	bitb	#RTSBIT
	beq	pirq1a
	andb	#RTSBIT^0xFF	; RTS:=1, LED on
	bra	pirq1b
pirq1a:	orb	#RTSBIT		; RTS:=0, LED off
pirq1b:	stb	ACIACS
	stb	*rtssts

  ;  change data direction of input ports to "READ"
  ; ------------------------------------------------

pirq2:	ldx	#RCRAM	       ; set pointer to base of array
	clr	*procin	       ; start at input port #0, I/O Board #0
	lda	*piadir	       ; set PIA port A direction to "INPUT"
	anda	#0xFF-DIRBIT
	sta	PIACA	       ; 1. access data direction reg A
	clra
	sta	PIADA	       ; 2. set port A == input
	lda	*piadir	       ; 3. now set buffer-74245 to input
	ora	#CA2BIT+DIRBIT ;    (prevent short-circuit moment
	sta	PIACA	       ;     between buffer-74245 and PIA)
	sta	*piadir	       ; 3. set flag: port A set to "INPUT"

  ;  loop through all defined active input ports and debounce switch contacts
  ; --------------------------------------------------------------------------

pirq2a:	lda	SWTDBC,x	; get SwitchToggleDebounce for this port
	deca			; SwitchToggleDebounce--
	bne	pirq2c
				; if SwitchToggleDebounce == 0
	lda	#BOUNCE
	sta	SWTDBC,x	; reload SwitchToggleDebounce
	ldb	*procin		; get current processing port number
	orb	#0xFF-IOADDR-RDWR ; negate DIS3S control
;	andb	#0xFF-RDWR	  ; direction: READ: enable "read" LS138
				  ; read physical port data
	stb	PIADB		; > setup correct input latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; > enable LS138 decoders
	nop			; > hardware settle delay
	lda	PIADA		; > read switch/toggle selected input port
	orb	#DIS3S
	stb	PIADB		; > disable LS138 decoders

	cmpa	PREVIN,x	; if physical_data == previous_data
	bne	pirq2b
	sta	INPORT,x	; set new key_data
	bra	pirq3
				; else
pirq2b:	sta	PREVIN,x	; previous_data = physical_data
	lda	#BOUNCE		; reload SwitchDebounce
				; else
pirq2c:	sta	SWTDBC,x	; store updated SwitchToggleDebounce
pirq3:	ldb	*procin
	incb			; move to next input port
	stb	*procin
	inx
	cmpb	#CONIHI		; all input ports processed ?
	bls	pirq2a		; n - read next input port


  ;  process console switches / toggles
  ; ------------------------------------

  ; * 1. always read Switch Register and store in SR

	lda	*INBASE+RD8e_1	; read SR[0:3]
	ldb	*INBASE+RD8e_0	; read SR[4:11]
	anda	#SW_swh		; isolate SR[0:3]
	std	*SR		; set Switch Register


  ; * 2. check status of HALT and S/STEP keys // set <conact> flag if necessary
  ;
  ;   After power-up [conact] == $00, *any* state of the switches is possible!
  ;   HALT   = off  --> [conact] not changed, process all toggles
  ;			CONT activated  --> [conact] := ACTrun
  ;   S/STEP = off  --> [conact] not changed, process all toggles
  ;			CONT activated  --> [conact] := ACTrun
  ;   HALT   = on   --> [conact] := $00, process all toggles
  ;			CONT activated  --> [conact] := ACTstp
  ;   S/STEP = on   --> [conact] := $00, process all toggles
  ;			CONT activated  --> [conact] := ACTcyc

	lda	*INBASE+RD8e_1	; check if HALT or S/STEP switch is active
	anda	#SW_hlt+SW_stp	; note: in active state the bit is 0 !
	cmpa	#SW_hlt+SW_stp
	beq	procsw		; both are NOT active: [conact] not changed
	clr	*conact		; any of the two is active: [conact] := $00


  ; * 3. process CONTINUE toggle (action depends on state of HALT and S/STEP)

procsw:	ldb	*INBASE+RD8e_3	; read toggles
; 	bitb	#SW_cnt		; CONT active (bit is 1) ?
;	beq	15$		; n -
	bpl	15$		; n - (performance: CONT at bit 7)

  ;   # CONT active (bit is 1) - check state of HALT and S/STEP switches
  ;	   HALT = on	--> [conact] := ACTstp
  ;	   S/STEP = on	--> [conact] := ACTcyc
  ;	   both  off	--> [conact] := ACTrun

10$:	lda	#ACTrun		; assume both off: to "run" state
	sta	*conact
	ldb	*INBASE+RD8e_1	; read HALT and S/STEP switch
	lda	#ACTcyc		; [conact] value if S/STEP active (bit is 0)
	bitb	#SW_stp		;   (cycle instructions not yet implemented)
	beq	12$		;   (HALT is don't care if S/STEP is active)
11$:	bitb	#SW_hlt
	bne	15$
	lda	#ACTstp		; HALT active :: step instructions
12$:	sta	*conact
	bra	20$


  ; * 4. process other toggles (ADDR-LOAD, EXTD-ADDR-LOAD, EXAM, DEP, CLEAR)
  ;	 Activation of these is only honored if [conact] != ACTrun

15$:	lda	*conact		; accept toggles ? [ACTrun] - (bit7) is clear
	lbmi	99$		; n -

   ; ## process ADDRESS LOAD toggle

20$:	bitb	#SW_al		; ADDR-LOAD active (bit is 1) ?
	bne	24$		; y -
	lda	*dlyal		; n - ADDR-LOAD at least TFRDLY inactive ?
	beq	30$
	deca
	bra	29$
24$:	lda	*dlyal		; was ADDR-LOAD inactive long enough ?
	bne	28$		; n - reload wait time for ADDR-LOAD toggle
	ldd	*SR		; y - copy SR to Program Counter
	std	*PC
	ldb	*INBASE+RD8e_3
28$:	lda	#TFRDLY		; reload wait time for ADDR-LOAD toggle inactive
29$:	sta	*dlyal

   ; ## process EXTENDED ADDRESS LOAD toggle

30$:	bitb	#SW_eal		; EXTD-ADDR-LOAD active (bit is 1) ?
	bne	34$		; y -
	lda	*dlyeal		; n - EXTD-ADDR-LOAD at least TFRDLY inactive ?
	beq	40$
	deca
	bra	39$
34$:	lda	*dlyeal		; was EXTD-ADDR-LOAD inactive long enough ?
	bne	38$		; n - reload wait time for ADDR-LOAD toggle
	lda	*SR		; y - read SR4-SR11
	anda	#Dbits		;     --> D-field (SR9-10-11)
	sta	*WorkD
	sta	*SaveD
	lda	*SR		;     read SR4-SR11
	asra
	asra
	asra
	anda	#Fbits		;     --> I-field (SR6-7-8)
	sta	*savacc		;	  save I-field
	tst	*WorkUI
	bpl	31$
	ora	#Ubit		;	  preserve U bit
31$:	sta	*WorkUI
	lda	*savacc		;	  restore I-field
	tst	*SaveUI
	bpl	32$
	ora	#Ubit
32$:	sta	*SaveUI
	lda	*savacc		;	  restore I-field
	tst	*BuffUI
	bpl	33$
	ora	#Ubit
33$:	sta	*BuffUI
38$:	lda	#TFRDLY		; reload wait time for EXTD-ADDR-LOAD toggle inactive
39$:	sta	*dlyeal

   ; ## process EXAMINE toggle

40$:	bitb	#SW_exm		; EXAM active (bit is 1) ?
	bne	44$		; y -
	lda	*dlyexm		; n - EXAM at least TFRDLY inactive ?
	beq	50$
	deca
	bra	49$

   ; 1)	Get current 'active' address, NXM is allowed.
   ;    Since the emulator code loops at "Doinst" while HALT or S/STEP is active,
   ;    MD (Memory Data) holds automatically the correct information.
   ; 2)	Increment 'active' address.
   ; ** NOT (yet) CORRECT: first EXAMINE request increments PC ...

44$:	lda	*dlyexm		; was EXAM inactive long enough ?
	bne	48$		; n - reload wait time for EXAM toggle

   ; actually reading the data from [PC] and writing that data into [MD]
   ; is not necessary, as the main loop (Doinst) in the simulator does this!

;	lda	*WorkUI		; y - accept EXAM toggle - get I field
;	lsla
;	ldx	#GetFld
;	ldx	a,x		; start address in memory of this Field
;	bne	46$
;	ldd	#0		; NXM - Non eXistent Memory
;	std	*MD		; --> data read = 0000
;	bra	47$
; 46$:	ldd	*PC
;	lslb			; *2 :: memory -> core address in Field
;	rola
;	ldd	d,x		; get data from physical core address
;	std	*MD		; show in MD on console
47$:	ldd	*PC
	addd	#1		; auto-increment Program Counter
	std	*PC
	ldb	*INBASE+RD8e_3
48$:	lda	#TFRDLY		; reload wait time for EXAM toggle inactive
49$:	sta	*dlyexm

   ; ## process DEPOSIT toggle

50$:	bitb	#SW_dep		; DEP active (bit is 1) ?
	bne	54$		; y -
	lda	*dlydep		; n - DEP at least TFRDLY inactive ?
	beq	60$
	deca
	bra	59$

   ; 1)	Get current 'active' address, NXM is allowed.
   ; 2)	Copy data in [SR] to the 'active' address.
   ;    Current implementation: if address is NXM, deposit action not done.
   ;	??? The contents of [SR] is displayed in the pdp8 register MD (???)
   ; 3)	Increment 'active' address.

54$:	lda	*dlydep		; was DEP inactive long enough ?
	bne	58$		; n - reload wait time for DEP toggle
	lda	*WorkUI		; get I field
	lsla
	ldx	#GetFld
	ldx	a,x		; start address in memory of this Field
	beq	56$		; NXM - Non eXistent Memory: no deposit!
	ldd	*PC
	lslb			; *2 :: memory -> core address in Field
	rola
	leax	d,x		; physical core address
	ldd	*SR
	std	,x		; store data from SR
56$:	ldd	*PC
	addd	#1		; auto-increment Program Counter
	std	*PC
	ldb	*INBASE+RD8e_3
58$:	lda	#TFRDLY		; reload wait time for DEP toggle inactive
59$:	sta	*dlydep

   ; ## process CLEAR toggle

60$:	bitb	#SW_clr		; check CLEAR active (bit is 1)
	bne	64$		; y -
	lda	*dlyclr		; n - CLEAR at least TFRDLY inactive ?
	beq	99$
	deca
	bra	69$
64$:	lda	*dlyclr		; was CLEAR inactive long enough ?
	bne	68$		; n - reload wait time for CLEAR toggle
	jsr	Clear		; activated: reset the pdp8 environment
68$:	lda	#TFRDLY		; reload wait time for CLEAR toggle inactive
69$:	sta	*dlyclr


  ;  output current state to LED's
  ; -------------------------------

99$:	lda	*piadir		; set data output port == OUTPUT
	anda	#0xFF-CA2BIT
	sta	PIACA		; 1. only the 245 buffer (prevent short-circuit)
	sta	*piadir		;    set flag: port A set to "OUTPUT"
	anda	#0xFF-DIRBIT	; 2. access data direction reg A
	sta	PIACA
	lda	#0xFF
	sta	PIADA		; 3. set port A == output
	lda	*piadir
	sta	PIACA		; 4. access data output reg A
;
	lda	*WorkUI		; output the current (pdp8/e) address
	lsla
	lsla
	lsla
	lsla
	ora	*PC		; EMA[210] (?) , A[11-8]
	ldb	#WR8e_1		; output upper 7 bits
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

	lda	*PC+1		; A[7-0]
	ldb	#WR8e_0		; output lower 8 bits
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

    ; check the Indicator Selector Switch (ISS)

	lda	INBASE+RD8e_2	; read ISS bits - get correct data
	ldx	#rc_ste
	bita	#ISSste
	bne	_shw1		; -> STATE
	ldx	#AC
	bita	#ISSac
	bne	_shw1		; -> AC
	ldx	#MD
	bita	#ISSmd
	bne	_shw1		; -> MD
	ldx	#MQ
	bita	#ISSmq
	bne	_shw1		; -> MQ
	ldx	#rc_bus
	bita	#ISSbus
	bne	_shw1		; -> BUS
	bsr	ISSSTS		; must be STATUS (special handling)
_shw1:	ldd	,x

	stb	*FIRQsv		; save lower 8-bit data
	ldb	*conact		; status == ACTrun (bit 7 = "running") ?
	bpl	no_run
	adda	#RUNled		; add RUN status LED = ON

no_run:	ldb	#WR8e_3		; output upper 4 bits
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

	ldb	#WR8e_2		; output lower 8 bits
	lda	*FIRQsv
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

	puls	d,x
dumirq:	rti


	.else

;  getoct - get octal value
;
;    call   : [B] = number of octal digits (1 to 4)
;    return : V=1     - entry aborted
;		        [D] = invalid
;	      V=0 C=1 - *first* entry is CR (used to accept "default")
;		        [D] = invalid
;	      V=0 C=0 - valid entry, value in [D]

getoct:	clra
	tfr	d,x	; set loop counter for requested digits
	pshs	x	; save for 1st char == CR check
	clrb
	std	*readpc	; init result
getoc1:	jsr	inch	; get char, no echo
	cmpa	#CR
	beq	getocd	; - accept default if 1st entry
	cmpa	#CtrlC
	beq	getocc	; - ^C : abort entry
	cmpa	#'0
	blo	getoc1	; accept 0-7 only
	cmpa	#'7
	bhi	getoc1
	jsr	outch	; print entry
	suba	#'0	; base-0
	pshs	a	; put octal digit on stack
	ldd	*readpc	; get current value
	lslb		; make room for new digit
	rola
	lslb
	rola
	lslb
	rola
	orb	,s+	; add new digit from stack
	std	*readpc	; update result
	leax	-1,x
	bne	getoc1
getoce:	andcc	#0xFC	; V=0 C=0 - result value in [D]
	puls	x,pc	; clean up stack
;
getocd:	cmpx	,s	; was CR the first entry?
	bne	getoce	; n - return accumulated value
	orcc	#1	; y - V=0 C=1
	puls	x,pc
;
getocc:	orcc	#2	; V=1 - abort entry
	puls	x,pc

	.endif




;  * ISSSTS - build Indicator Selector Switch STATUS data
;
;    call   : ### also called from FIRQ routine if ISS at "STATUS" ###
;    return : [X] - ptr to collected bit mask (word)
;    note   : Greater Than flag not (yet) implemented
;
;    ----------------------------------------------------------------
;   | LNK  GT   INT   NO  ION   UM  IF0  IF1   IF2  DF0   DF1   DF2  |
;   |           BUS  INT                                             |
;   |=====----===========+===========================================|
;   |  0    1    2    3    4    5    6    7    8    9    10     11   |
;    --------------------+-------------------------------------------
;               note: an "=" means: bit mentioned above it is in code!

ISSSTS:	clra
	ldb	*Link
	bitb	#Lmask
	beq	stshi2
	lda	#LD_lnk		; Link
stshi2:				; *** Greater Than flag ***
stshi1:	ldb	*IOflgs
	andb	#Jdelay^0xFF
	orb	*IOflgs+1
	beq	stshi0
	ora	#LD_irb		; Interrupt Bus
stshi0:	ldb	*IOflgs
	andb	#Jdelay
	beq	stslo5
	ora	#LD_nia		; No Interrupt Allowed
stslo5:	sta	*issts2
	ldb	*WorkUI
	andb	#Fbits
	lslb
	lslb
	lslb			; Instruction Field 0-1-2
stslo7:	ldx	*emufnc
	cmpx	#Chkint
	bne	stslo6
	orb	#LD_ion		; Interrupt On
stslo6:	lda	*WorkUI
	bpl	stslo2
	orb	#LD_umd		; User Mode
stslo2:	lda	*WorkD
	anda	#Dbits
	sta	*issts2+1
	addb	*issts2+1	; Data Field 0-1-2
	stb	*issts2+1
	ldx	#issts2
	rts



	.if	RealCons
	.else

;  ShwReg - show all register information (incl. ASCII artwork)
;
;    call   : [A] = requested register to display
;    return : print "lights" based on ISS selection

ShwReg:	ldx	#line1
	jsr	pdatao
	cmpa	#ISSste		; check selected display line
	bne	ISS_5
;     ______________________________________________________________
;    |  F    D    E   IR0  IR1  IR2  MD   DATA  SW PAUSE  BRK  BRK  |
;    |                               DIR  CONT            PROG      |
;    |==============================================================|
;    |  ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   | STATE
;    ----------------------------------------------------------------

ISS_6:	ldd	*rc_ste		; get STATE data
	ldy	#stenam		;     and ID,
	ldx	#stelin		;     1st line STATE format, ASCII art
	bra	ISSDPY

ISS_5:	cmpa	#ISSsts
	bne	ISS_4
;     ______________________________________________________________
;    | LNK   GT  INT   NO  ION   UM  IF0  IF1  IF2  DF0  DF1  DF2   |
;    |           BUS  INT                                           |
;    |==============================================================|
;    |  ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   | STATUS
;    ----------------------------------------------------------------

	bsr	ISSSTS		; collect STATUS flag bits
	ldd	,x		; load the flag bits in [D]
	ldy	#stsnam
	ldx	#stslin
	bra	ISSDPY
;     ______________________________________________________________
;    |  0    1    2    3    4    5    6    7    8    9    10   11   |
;    |==============================================================|
;    |  ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   | AC/MD/MQ/BUS
;    ----------------------------------------------------------------

ISS_4:	ldx	#numlin
	cmpa	#ISSbus
	bne	ISS_3
	ldd	#0		; "BUS" (to be implemented ...)
	ldy	#busnam
	bra	ISSDPY

ISS_3:	cmpa	#ISSmd
	bne	ISS_2
	ldd	*MD
	ldy	#mdnam
	bra	ISSDPY

ISS_2:	cmpa	#ISSmq
	bne	ISS_1
	ldd	*MQ
	ldy	#mqnam
	bra	ISSDPY

ISS_1:	ldd	*AC		; must be "AC"
	ldy	#acnam

   ; [D] - light data
   ;       for every bit [11-0] set "##" printed, else "  "
   ; [Y] - register ID text string
   ; [X] - pointer to 1st line format, ASCII art

ISSDPY:	pshs	d
	jsr	pdatao		; print register ASCII art line(s)
	lda	#'|
	jsr	outch
	lda	#'=
	ldb	#62
	jsr	OUTABT		; lights separator line
	ldx	#line2
	jsr	pdatao		; goto "lights" line
	puls	d
	pshs	b
	lsla			; move to 1st bit in msb
	lsla
	lsla
	lsla
	ldb	#4		; print 4 bits [11-8]
	bsr	shwlgt		; show the 4 "lights"
	puls	a
	ldb	#8		; print 8 bits [7-0]
	bsr	shwlgt
	tfr	y,x
	jsr	pdatao		; print register name
	jsr	docrlf
	lda	#'-
	ldb	#64
	jsr	OUTABT		; lights separator line
	jmp	docrlf



;  * shwlgt - show "lights"
;
;    call   : [A] = "light" bits
;	      [B] = # lights to print
;    return : [A], [B], [X] destroyed

shwlgt:	ldx	#OFF2sp		; assume bit = clear
	lsla
	bcc	shwlg1
	ldx	#ON3sp		; nope,  bit = set
shwlg1:	jsr	pdatao
	decb
	bne	shwlgt
	rts



   ; ASCII "art" register display

line1:	.fcb	CR,LF
	.fcc	" ______________________________________________________________"
	.fcb	CR,LF,0
stenam:	.asciz	"| STATE"
stelin:	.fcc	"|  F    D    E   IR0  IR1  IR2  MD   DATA  SW PAUSE  BRK  BRK  |"
	.fcb	CR,LF
	.fcc	"|                               DIR  CONT            PROG      |"
	.fcb	CR,LF,0
stsnam:	.asciz	"| STATUS"
stslin:	.fcc	"| LNK   GT  INT   NO  ION   UM  IF0  IF1  IF2  DF0  DF1  DF2   |"
	.fcb	CR,LF
	.fcc	"|           BUS  INT                                           |"
	.fcb	CR,LF,0
numlin:	.fcc	"|  0    1    2    3    4    5    6    7    8    9    10   11   |"
	.fcb	CR,LF,0
line2:	.fcc	"|"
	.fcb	CR,LF
	.asciz	"|  "
acnam:	.asciz	"| AC"
mdnam:	.asciz	"| MD"
mqnam:	.asciz	"| MQ"
busnam:	.asciz	"| BUS"
ON3sp:	.asciz	"##   "
OFF2sp:	.asciz	"     "

SSmsg:	.fcb	CR,LF,CR,LF
	.fcc	"   <SPC>:single step   <CR>:run   <K>:keyboard input"
	.fcb	CR,LF
	.fcc	" <1>:STATE  <2>:STATUS  <3>:AC  <4>:MD  <5>:MQ  <6>:BUS"
	.fcb	CR,LF,CR,LF,0

Kbdmsg:	.fcb	CR,LF
	.asciz	"keyboard entry: "

askpc:	.fcb	CR,LF
	.asciz	"enter start PC [0200] 4-digit octal ->"

	.endif




   ; simulator messages

Sysmsg:	.fcb	CR,LF
	.fcc	"*** pdp8/e  SIMULATOR  v0.3 ***"
	.fcb	CR,LF,0

ldrmss:	.fcb	CR,LF
	.asciz	"download binary file -> end with ^Q ..."
ldrmse:	.fcc	" completed succesfully"
	.fcb	CR,LF
	.asciz	"load another file [Y/N] : "
Lderr:	.asciz	"Error loading binary file"
ldrcnt:	.fcb	CR,LF
	.asciz	"! at file byte count = $"
Errfld:	.asciz	"Field specification invalid"
Errmem:	.asciz	"Field (memory) does not exist"
Errcks:	.asciz	"Checksum failed"
;NoXON:	.asciz	"No XON received as a start character after load"

TTO_3_4:	.asciz	"IOP #3 or #7 attempted on device 04 (TTO)"
Insterr:	.asciz	"Illegal instruction"
KM8Eerr:	.asciz	"Unimplemented KM8-E instruction"
TTI_3_7:	.asciz	"IOP #3 or #7 attempted on device 03 (TTI)"

Clrmsg:	.asciz	"Error in Clear subroutine (from IOset call)"
Hltmsg:	.asciz	"Program halt (HLT microcode in OPR instruction)"
Bitmsg:	.asciz	"Bits set in locations that should be 0"
NoIerr:	.asciz	"Unimplemented interrupt IOP attempted (IOT-00 #6)"
NoFLD:	.asciz	"No memory allocated for this field"

	.end
